import os
import subprocess
from typing import List, Optional

# The HTML rendering is expected to be a snippet of an HTML document as opposed to a standalone webpage.
# The following will ignore validation errors that are meant for a standalone HTML document.
_WHITELIST_ERRORS: List[str] = [
    # DOCTYPE tag is expected at the top of the HTML document; ignore in snippet.
    r'Expected "<!DOCTYPE html>"',
    # Title element defines the documents title shown in the browser; ignore in snippet.
    r'Element "head" is missing a required instance of child element "title"',
    r"Non-space character in page trailer",
    # Some CSS background-color hex values are reported as invalid.
    r'not a "background-color" value',
    # The following are due to multiple HTML <html> elements in a single snippet.
    # Valid HTML has one per document, but we render two:
    # 1) Riptable object and styles
    # 2) Metadata about the Riptable object that was rendered such as shape and bytes
    r'Stray start tag "html"',
    r'Stray start tag "p"',
    r"fatal: Cannot recover after last error",
]


class VNUChecker:
    _CN = "VNUChecker"
    _JAVA = r"java"
    _test_dispaly_path = os.path.join(
        os.getcwd(),
        r"Python",
        r"core",
        r"riptable",
        r"test_tooling_integration",
        r"test_display",
    )
    _JAR_PATH = os.path.join(_test_dispaly_path, r"vnu_jar", r"vnu.jar")
    _BASE_PATH = os.path.join(_test_dispaly_path, r"test_display", r"html_output")

    def __init__(
        self,
        java: Optional[str] = None,
        jar_path: Optional[str] = None,
        base_path: Optional[str] = None,
        errors_only: bool = False,
        ascii_quotes: bool = False,
    ):
        self._java = java
        self._jar_path = jar_path
        self.base_path = base_path
        self._errors_only = errors_only
        self._ascii_quotes = ascii_quotes

        if self._java is None:
            self._java = VNUChecker._JAVA
        if self._jar_path is None:
            self._jar_path = VNUChecker._JAR_PATH
        if self.base_path is None:
            self.base_path = VNUChecker._BASE_PATH

        self._args = self._build_args()

    def __str__(self):
        return " ".join(self._args)  # The command line representation.

    def __repr_(self):
        return f"{VNUChecker._CN}(java={self._java}, jar_path={self._jar_path}, dir_path={self.base_path}, errors_only={self._errors_only}, ascii_quotes={self._ascii_quotes})"

    def _is_whitelist_error(self, error_text: str, extra_whitelists: Optional[List[str]] = None) -> bool:
        """Returns ``False`` if ``error_text`` is a whitelisted error, otherwise ``True``."""
        if extra_whitelists is None:
            extra_whitelists = []
        whitelist_errors = _WHITELIST_ERRORS.copy() + extra_whitelists
        for we in whitelist_errors:
            if error_text.find(we) != -1:  # found whitelisted error
                return False
        return True

    def _build_args(self) -> List[str]:
        """Returns a list of program arguments that are used to kick of the VNU Checker."""
        cmd: List[str] = [self._java, "-jar", self._jar_path]
        if self._ascii_quotes:
            cmd.append("--asciiquotes")
        if self._errors_only:
            cmd.append("--errors-only")
        cmd.append(self.base_path)
        return cmd

    def _run(self, args: Optional[List[str]] = None) -> List[str]:
        """Runs the VNU Checker and returns a list of errors that are not whitelisted.
        Uses the default program arguments if none are specified.
        """
        if args is None:
            args = self._args
        p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        _, stderr = p.communicate()
        errors = stderr.decode("utf-8").splitlines()
        return list(filter(self._is_whitelist_error, errors))

    def validate(self, dir_path: Optional[str] = None) -> List[str]:
        """``validate`` will run the VNU Checker and return a list of errors. An empty list signals no errors.
        ``validate`` takes an optional directory path to run the VNU Checker, otherwise the default one is used.
        """
        if dir_path is None:
            if self.base_path is None:
                raise ValueError(f"{self._CN}.validate: need a directory to validate")
            dir_path = self.base_path
        if not os.listdir(dir_path):
            raise ValueError(f"{self._CN}.validate: {dir_path} is empty; no files to validate")
        return self._run()



__all__ = [
    "build_header_tuples",
    "output_cache_flush",
    "output_cache_none",
    "output_cache_setsize",
    "parse_header_tuples",
    "profile_func",
    "sub2ind",
    "autocomplete",
    "jedi_completions",
]

import warnings

import numpy as np

from .rt_enum import ColHeader, TypeRegister

#
#
# MATLAB
# sub2ind([10 10], 3,3)
# ans = 23.00
#
# ------------------------
# python
# In [10]: np.ravel_multi_index((3,3),(10,10))
# Out[10]: 33
#
# In [6]: np.unravel_index(23,(10,10))
# Out[6]: (2, 3)
#
# ----------------------------------------------
# MATLAB
# sub2ind([7 13], 6, 12)
# ans =  83.00
#
# sub2ind([7 13], 5, 12)
# ans =  82.00
#
# sub2ind([7 15], 5, 12)
# ans =  82.00
#
# np.ravel_multi_index((11,5),(13,7))
# 82
#
# np.ravel_multi_index((5,11),(7, 13), order='F')
# 82
# ALSO
# np.ravel_multi_index((5,11),(7, 15), order='F')
# 82

# MATH for this.. where is 5,11 in the array (7,13) if the array (7,13) was flattened
# 5, 11 in Fortran order means
# each row is 11 items and there are 5 columns
#
# so each row 11*7 = 77 + 5 = 82
#


def sub2ind(aSize, aPosition):
    """
    **MATLAB**

    sub2ind Linear index from multiple subscripts.
    sub2ind is used to determine the equivalent single index
    corresponding to a given set of subscript values.

    IND = sub2ind(SIZ,I,J) returns the linear index equivalent to the
    row and column subscripts in the arrays I and J for a matrix of
    size SIZ.

    IND = sub2ind(SIZ,I1,I2,...,IN) returns the linear index
    equivalent to the N subscripts in the arrays I1,I2,...,IN for an
    array of size SIZ.

    I1,I2,...,IN must have the same size, and IND will have the same size
    as I1,I2,...,IN. For an array A, if IND = sub2ind(SIZE(A),I1,...,IN)),
    then A(IND(k))=A(I1(k),...,IN(k)) for all k.

    **PYTHON**

    ravel_multi_index(...)
    ravel_multi_index(multi_index, dims, mode='raise', order='C')

    Converts a tuple of index arrays into an array of flat
    indices, applying boundary modes to the multi-index.

    Parameters
    ----------
    multi_index : tuple of array_like
        A tuple of integer arrays, one array for each dimension.
    dims : tuple of ints
        The shape of array into which the indices from ``multi_index`` apply.
    mode : {'raise', 'wrap', 'clip'}, optional
        Specifies how out-of-bounds indices are handled.  Can specify
        either one mode or a tuple of modes, one mode per index.

        * 'raise' -- raise an error (default)
        * 'wrap' -- wrap around
        * 'clip' -- clip to the range

        In 'clip' mode, a negative index which would normally
        wrap will clip to 0 instead.
    order : {'C', 'F'}, optional
        Determines whether the multi-index should be viewed as
        indexing in row-major (C-style) or column-major
        (Fortran-style) order.

    Returns
    -------
    raveled_indices : ndarray
        An array of indices into the flattened version of an array
        of dimensions ``dims``.

    See Also
    --------
    unravel_index

    Notes
    -----
    .. versionadded:: 1.6.0

    Examples
    --------
    >>> arr = np.array([[3,6,6],[4,5,1]])
    >>> np.ravel_multi_index(arr, (7,6))
    array([22, 41, 37])
    >>> np.ravel_multi_index(arr, (7,6), order='F')
    array([31, 41, 13])
    >>> np.ravel_multi_index(arr, (4,6), mode='clip')
    array([22, 23, 19])
    >>> np.ravel_multi_index(arr, (4,4), mode=('clip','wrap'))
    array([12, 13, 13])

    >>> np.ravel_multi_index((3,1,4,1), (6,7,8,9))
    1621
    """
    return np.ravel_multi_index(aPosition, aSize, order="F")


# ---------------------------------------------------------------------------
def build_header_tuples(headers, span, group):
    if headers is None:
        return []
    return [ColHeader(name, span, group) for name in headers]


# ---------------------------------------------------------------------------
def parse_header_tuples(header_tups):
    return [h.col_name for h in header_tups]


# ----------------------------------------------
def jedi_completions(text, offset):
    """
    autocomplete() must be called first.
    Not used yet. Returns the same completions jedi would.

    Examples
    --------
    from riptable.rt_misc import jedi_completions
    st = Struct({'a': 5})
    jedi_completions('st', 2)
    """

    def position_to_cursor(text: str, offset: int):
        before = text[:offset]
        blines = before.split("\n")
        line = before.count("\n")
        col = len(blines[-1])
        return line, col

    def cursor_to_position(text: str, line: int, column: int) -> int:
        lines = text.split("\n")
        return sum(len(l) + 1 for l in lines[:line]) + column

    try:
        ipc = Hooker._ipcompleter
        cursor_line, cursor_column = position_to_cursor(text, offset)

        namespaces = [ipc.namespace]
        if ipc.global_namespace is not None:
            namespaces.append(ipc.global_namespace)

        completion_filter = lambda x: x
        offset = cursor_to_position(text, cursor_line, cursor_column)
        # filter output if we are completing for object members
        if offset:
            pre = text[offset - 1]
        #            if pre == '.':
        #                if self.omit__names == 2:
        #                    completion_filter = lambda c:not c.name.startswith('_')
        #                elif self.omit__names == 1:
        #                    completion_filter = lambda c:not (c.name.startswith('__') and c.name.endswith('__'))
        #                elif self.omit__names == 0:
        #                    completion_filter = lambda x:x
        #                else:
        #                    raise ValueError("Don't understand self.omit__names == {}".format(self.omit__names))

        import jedi

        interpreter = jedi.Interpreter(text[:offset], namespaces, column=cursor_column, line=cursor_line + 1)
        return interpreter.completions()

    except Exception:
        return []


class CacheWarning(UserWarning):
    pass


# ------------------------------------------------------------------------
def output_cache_none():
    """
    used in ipython, jupyter, or spyder
    sets the terminalInteractiveShell output cache size to none
    Out[#] will no longer work
    the Out dictionary will be empty
    _# will no longer work
    """
    try:
        import IPython

        t = IPython.terminal.interactiveshell.TerminalInteractiveShell()
        # stop the interactiveshell from caching (this is the Out[])
        t.cache_size = 0
        # note: do we set the use_ns Out back to something to allow to kick back in?
        # stop the displayhook from caching
        ipython = IPython.get_ipython()
        if ipython:
            ipython.displayhook.do_full_cache = False
    except:
        pass
        # warnings.warn("Failed to set output_cache_none.", CacheWarning)


# ------------------------------------------------------------------------
def output_cache_setsize(cache_size=100):
    """
    used in ipython, jupyter, or spyder
    sets the terminalInteractiveShell output cache size to cache_size (100 is the default)
    """
    try:
        import IPython

        t = IPython.terminal.interactiveshell.TerminalInteractiveShell()
        # stop the interactiveshell from caching (this is the Out[])
        t.cache_size = cache_size
        # stop the displayhook from caching
        ipython = IPython.get_ipython()
        if ipython:
            ipython.displayhook.do_full_cache = True
        # note: consider setting cull_fraction or cache_size
    except:
        warnings.warn("Failed to set output_cache_setsize.", CacheWarning)


# ------------------------------------------------------------------------
def output_cache_flush():
    """
    used in ipython, jupyter, or spyder
    calling output_cache_flush() will remove object reference in the output cache
    it is recommended this is called when there are memory concerns
    """
    try:
        from IPython import get_ipython

        # from IPython import InteractiveShell
        ipython = get_ipython()
        if not ipython:
            return
        tempcache = ipython.displayhook.do_full_cache
        ipython.displayhook.do_full_cache = True
        ipython.displayhook.flush()
        # ipython.displayhook.cache_size =0
        del ipython.displayhook.shell.user_ns["_"]
        shell = ipython
        # ns_refs = [m.__dict__ for m in shell._main_mod_cache.values()]
        ## Also check in output history
        # ns_refs.append(shell.history_manager.output_hist)
        # for ns in ns_refs:
        #    to_delete = [n for n, o in ns.items()]
        #    for name in to_delete:
        #        if (ns[name] is obj):
        #            print("**deleting", name)
        #            del ns[name]

        ns_refs = [shell.user_ns, shell.user_global_ns, shell.user_ns_hidden]
        for ns in ns_refs:
            to_delete = [n for n, o in ns.items()]
            for name in to_delete:
                if name in ("_", "__", "___"):
                    print("**deleting", name, type(name))
                    del ns[name]

                # if ns[name] is obj:
                #    print("**extra deleting", name)
                #    del ns[name]

                # search for variables named _# such as _4 or _37
                elif name.startswith("_") and name[1] >= "0" and name[1] <= "9":
                    temp = ns[name]

                    # TODO: consider looking for more than just Dataset
                    if isinstance(temp, TypeRegister.Dataset):
                        print("**extra deleting Dataset", name, "size of ", temp._last_row_stats())
                    elif isinstance(temp, np.ndarray):
                        print("**extra deleting array", name, "size of ", (temp.itemsize * temp.size) / 1e6, "MB")
                    else:
                        print("**extra deleting array", name, "type ", type(temp))
                    del ns[name]

        # Ensure it is removed from the last execution result
        shell.last_execution_result = None

        # displayhook keeps extra references, but not in a dictionary
        for name in ("_", "__", "___"):
            setattr(shell.displayhook, name, None)

        # set it back to what it was
        ipython.displayhook.do_full_cache = tempcache

    except:
        warnings.warn("Failed to set output_cache_flush.", CacheWarning)


# ------------------------------------------------------------------------
def profile_func(func, sortby="time"):
    """
    Used to profile a function that has no arguments

    Examples
    --------
    This will time how long the __repr__ function to print out a dataset

    >>> import riptable_docdata as rtd
    >>> trips = rt.Dataset(rtd.get_bike_trips_data('trips'))
    >>> profile_func(trips.__repr__)
    """

    try:
        import cProfile

        pr = cProfile.Profile()
        pr.enable()

        func()

        pr.disable()
        pr.print_stats(sort=sortby)
    except:
        print(f"cProfile could not be imported or could not run {func}")


# simple class to hold what we hooked
class Hooker:
    _ipcompleter = None
    _orig_do_complete = None
    _orig_deduplicate = None
    _putils = None
    babydict = {c: None for c in list("_abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.")}


# The Completion class is copied
class Completion:
    """
    Completion object used and return by IPython completers.

    .. warning:: Unstable

        This function is unstable, API may change without warning.
        It will also raise unless use in proper context manager.

    This act as a middle ground :any:`Completion` object between the
    :any:`jedi.api.classes.Completion` object and the Prompt Toolkit completion
    object. While Jedi need a lot of information about evaluator and how the
    code should be ran/inspected, PromptToolkit (and other frontend) mostly
    need user facing information.

    - Which range should be replaced replaced by what.
    - Some metadata (like completion type), or meta information to displayed to
      the use user.

    For debugging purpose we can also store the origin of the completion (``jedi``,
    ``IPython.python_matches``, ``IPython.magics_matches``...).
    """

    __slots__ = ["start", "end", "text", "type", "signature", "_origin"]

    def __init__(self, start: int, end: int, text: str, *, type: str = None, _origin="", signature="") -> None:

        self.start = start
        self.end = end
        self.text = text
        self.type = type
        self.signature = signature
        self._origin = _origin

    def __repr__(self):
        return "<Completion start=%s end=%s text=%r type=%r, signature=%r,>" % (
            self.start,
            self.end,
            self.text,
            self.type or "?",
            self.signature or "?",
        )

    def __eq__(self, other):
        """
        Equality and hash do not hash the type (as some completer may not be
        able to infer the type), but are use to (partially) de-duplicate
        completion.

        Completely de-duplicating completion is a bit tricker that just
        comparing as it depends on surrounding text, which Completions are not
        aware of.
        """
        return self.start == other.start and self.end == other.end and self.text == other.text

    def __hash__(self):
        return hash((self.start, self.end, self.text))


def autocomplete(hook: bool = True, jedi: bool = None, greedy: bool = None):
    """
    Call rt.autocomplete() to specialize jupyter lab autcomplete output.
    arrays, categoricals, datetime, struct, and datasets will be detected.
    array will be array followed by the dtype.

    Parameters
    ----------
    hook: bool, default True
        set to False to unhook riptable autocomplete
    jedi: bool, default None
        set to True to set use_jedi in IPython
    greedy: bool, default None
        set to True to set greedy in IPython for [' autocomplete

    Examples
    --------
    >>> rt.autocomplete(); ds=Dataset({'test':arange(5), 'another':arange(5.0), 'mycat':rt.Cat(arange(5)), 'mystr': arange(5).astype('S')})
    Now in jupyter lab type 'ds.<tab>'

    """

    def gettype(element) -> str:
        result: str = ""
        if hasattr(element, "_autocomplete"):
            result = element._autocomplete()
        elif isinstance(element, np.ndarray):
            # how to display array
            bstring = "Array "
            dnum = element.dtype.num
            if dnum == 0:
                # bool check
                extra = "b"
            elif dnum <= 10:
                # integer check
                if dnum & 1 == 1:
                    extra = "i"
                else:
                    extra = "u"
            elif dnum <= 13:
                extra = "f"
            else:
                extra = element.dtype.char + str(element.itemsize)

            if dnum <= 13:
                result = bstring + extra + str(element.itemsize * 8)
            else:
                result = bstring + extra

        else:
            if callable(element):
                result = "function"
            else:
                try:
                    result = element.__class__.__name__
                except Exception:
                    result = "unknown"
        return result

    # ----------------------------------------------
    def gettype_foritem(acobj, name, oclass, oinstance) -> str:
        t = "<unknown>"
        try:
            item = None
            if name in oinstance:
                item = oinstance[name]
            elif name in oclass:
                item = oclass[name]
            elif hasattr(acobj, name):
                item = getattr(acobj, name)
            t = gettype(item)
        except Exception:
            pass
        return t

    # ---------------------------------------------
    def _evaluate_text(code, dotpos, careful: bool = False):
        """
        careful: when True indicates that the entire line should NOT be evaluated

        NOTE: Internal function that calls eval() which may not be acceptable for some applications.
        """

        def call_eval(text):
            try:
                # try two diff namespaces
                acobj = eval(mainpart, _ipcompleter.namespace)
            except:
                try:
                    acobj = eval(mainpart, _ipcompleter.global_namespace)
                except:
                    acobj = None
            return acobj

        mainpart = code[:dotpos]
        acobj = None
        startpos = dotpos
        endpos = dotpos + 1
        startwith = None

        if careful:
            # check for ()
            pass

        if not careful:
            acobj = call_eval(mainpart)

        if acobj is None:
            # trickier... now scan backwards from dot
            while startpos >= 0:
                # search back until hit non-naming character
                if code[startpos] not in Hooker.babydict:
                    startpos += 1
                    break
                startpos -= 1

            # careful did not check entire line
            if careful and startpos < 0:
                startpos = 0
            if startpos >= 0:
                mainpart = code[startpos:dotpos]
                acobj = call_eval(mainpart)

            # if we still have not found it
            if acobj is None:
                laststartpos = startpos
                startpos = dotpos
                while startpos >= 0:
                    # search back until hit paren() or comma
                    if code[startpos] in "(),":
                        startpos += 1
                        break
                    startpos -= 1

                # careful did not check entire line
                if careful and startpos < 0:
                    startpos = 0

                if startpos != laststartpos:
                    if startpos >= 0:
                        mainpart = code[startpos:dotpos]
                        acobj = call_eval(mainpart)

        if acobj is not None:
            # calc the subsection
            while endpos < len(code):
                # search back until hit non-naming character
                if code[endpos] not in Hooker.babydict:
                    break
                endpos += 1

            if endpos > (dotpos + 1):
                startwith = code[dotpos + 1 : endpos]
            # print("startpos", startpos, endpos, mainpart, startwith)
            startpos = dotpos

        return acobj, startpos, mainpart, endpos, startwith

    # ---------------------------------------------
    def _riptable_deduplicate_completions(text, completions):
        # This is the hook for use_jedi=True and console.
        # [<Completion start=3 end=3 text='as_struct' type='function', signature='(self)',>,  ..]
        # there is also special code to detect a 'function' in IPython\terminal\ptutils.py
        found = False

        # turn the enumerator into a list
        completions = list(completions)

        # look for apply_schema as only riptable will have this marker up front
        for comp in completions:
            if comp.text == "apply_schema" or comp.text == "apply_cols":
                found = True  # looks like our Struct
                break

        # if jedi completed something that is not a Struct, stop trying to autocomplete
        if not found and len(completions) > 0:
            return Hooker._orig_deduplicate(text, completions)

        # we only autocomplete on dots
        dotpos = text.rfind(".")

        if dotpos > 0:
            acobj, startpos, mainpart, endpos, startwith = _evaluate_text(text, dotpos, careful=not found)

            if acobj is not None:

                # calc the subsection
                endpos = dotpos + 1
                while endpos < len(text):
                    # search back until hit non-naming character
                    if text[endpos] not in Hooker.babydict:
                        break
                    endpos += 1

                # is this a container class we own
                if isinstance(acobj, (TypeRegister.Struct, TypeRegister.FastArray)):
                    oclass = acobj.__class__.__dict__
                    oinstance = acobj.__dict__

                    # TODO: if jedi has a mistake, do we correct it here?
                    if len(completions) == 0:
                        # redo completions
                        completions = []
                        ldir = dir(acobj)
                        for c in ldir:
                            if not c.startswith("_"):
                                # check if we have a startswith
                                if startwith is not None:
                                    if not c.startswith(startwith):
                                        continue

                                t = gettype_foritem(acobj, c, oclass, oinstance)
                                completions.append(
                                    Completion(start=startpos + 1, end=dotpos, text=c, type=t, signature="(self)")
                                )

                    if isinstance(acobj, TypeRegister.Struct):
                        # get the columns
                        keys = acobj.keys()
                        keys.sort()
                        movetotop = {}

                        # first put the columns in
                        for k in keys:
                            # check if we have a startswith
                            if startwith is not None:
                                if not k.startswith(startwith):
                                    continue

                            # Struct or Dataset getitem call
                            element = acobj[k]
                            movetotop[k] = Completion(
                                start=startpos + 1, end=dotpos, text=k, type=gettype(acobj[k]), signature="(self)"
                            )

                        # then add anything else (note if completions is empty we could call dir)
                        for comp in completions:
                            text = comp.text
                            if text is not None and text not in movetotop:
                                movetotop[text] = comp

                        completions = list(movetotop.values())

        return Hooker._orig_deduplicate(text, completions)

    def _riptable_do_complete(self, code, cursor_pos):
        """
        Hooked from ipythonkernel.do_complete.  Hit in jupyter lab.
        Calls the original do_complete, then possibly rearranges the list.
        As of Dec 2019, this is the use_jedi=True hook in jupyter lab.
        """

        # self is ipkernel.ipythonkernel
        # code is what text the user typed
        # call original first (usually kicks in jedi)
        result = Hooker._orig_do_complete(self, code, cursor_pos)

        # we only autocomplete on dots
        dotpos = code.rfind(".")

        if dotpos > 0:
            mainpart = code[:dotpos]
            acobj, startpos, mainpart, endpos, startwith = _evaluate_text(code, dotpos)

            if acobj is not None:
                try:
                    if isinstance(acobj, TypeRegister.Struct):

                        oclass = acobj.__class__.__dict__
                        oinstance = acobj.__dict__

                        # add a dot to complete mainpart for string matching later
                        mainpart += "."
                        lenmainpart = len(mainpart)

                        # get the jedi completions
                        rmatches = result["matches"]

                        # check if there are any jedi completions
                        if len(rmatches) == 0:

                            # jedi failed, so we will attempt
                            matches = []
                            completions = []
                            ldir = dir(acobj)
                            for c in ldir:
                                if not c.startswith("_"):
                                    # check if we have a startswith
                                    if startwith is not None:
                                        if not c.startswith(startwith):
                                            continue
                                    matches.append(c)
                                    t = gettype_foritem(acobj, c, oclass, oinstance)
                                    completions.append({"start": startpos, "end": dotpos, "text": c, "type": t})

                            rmatches = matches

                            result = {
                                "matches": matches,
                                "cursor_end": endpos,
                                "cursor_start": dotpos + 1,
                                "metadata": {"_jupyter_types_experimental": completions},
                                "status": "ok",
                            }

                        meta = result["metadata"]
                        keys = acobj.keys()
                        keys.sort()

                        # result will look similar to:
                        # {'matches': ['list of strings'], 'cursor_end': 5, 'cursor_start': 5, 'metadata': {'_jupyter_types_experimental':
                        #     [{'start': 5, 'end': 5, 'text': 'add_traits', 'type': 'function'},
                        #      {'start': 5, 'end': 5, 'text': 'update_config', 'type': 'function'}]}, 'status':'ok'}
                        # jupyter notebook
                        jtypes = meta.get("_jupyter_types_experimental", None)
                        if jtypes is not None:
                            toptext = []
                            bottomtext = []

                            topmatch = []
                            bottommatch = []

                            # jtypes is a list of dicts
                            for jdict in jtypes:
                                top = False
                                text = jdict.get("text", None)
                                if text is not None:
                                    subtext = text
                                    if text.startswith(mainpart):
                                        subtext = text[lenmainpart:]
                                    # jedi matches after the dot
                                    if subtext in keys:
                                        # Struct or Dataset getitem call
                                        element = acobj[subtext]
                                        jdict["type"] = gettype(element)
                                        top = True
                                if top:
                                    topmatch.append(jdict)
                                    toptext.append(text)
                                else:
                                    bottommatch.append(jdict)
                                    bottomtext.append(text)

                            # change the order
                            jtypes = topmatch + bottommatch
                            rmatches = toptext + bottomtext

                            # insert our order
                            meta["_jupyter_types_experimental"] = jtypes
                            result["matches"] = rmatches

                            if len(toptext) > 0 and dotpos < len(code):
                                # indicate end of list in some instances to force regen
                                # to move out list back to the top during partial completions like ds.a<tab>
                                msg = "---"
                                jtypes.append({"start": 0, "end": 0, "text": msg, "type": "endlist"})
                                rmatches.append(msg)

                except Exception as e:
                    # indicate we crashed so user can report
                    jtypes = meta.get("_jupyter_types_experimental", None)
                    msg = "CRASHRIPTABLE"
                    jtypes.insert(0, {"start": 0, "end": 0, "text": msg, "type": f"{e}"})
                    result["matches"].insert(0, msg)

        return result

    # ---- start of main code for autocomplete --------
    import IPython
    from ipykernel import ipkernel

    _ipcompleter = IPython.get_ipython().Completer
    Hooker._ipcompleter = _ipcompleter

    # check version, we only support 7
    version = IPython.version_info[0]
    if version != 7:
        return

    # caller may optionally set jedi or greedy
    if jedi is True or jedi is False:
        _ipcompleter.use_jedi = jedi
    if greedy is True or greedy is False:
        _ipcompleter.greedy = greedy

    if hook:
        # for jupyter lab
        if Hooker._orig_do_complete is None:
            Hooker._orig_do_complete = ipkernel.IPythonKernel.do_complete
            setattr(ipkernel.IPythonKernel, "do_complete", _riptable_do_complete)

        # for console text (not jupyter lab)
        if Hooker._orig_deduplicate is None:
            Hooker._putils = IPython.terminal.ptutils
            Hooker._orig_deduplicate = Hooker._putils._deduplicate_completions
            setattr(Hooker._putils, "_deduplicate_completions", _riptable_deduplicate_completions)
    else:
        # unhook ------
        if Hooker._orig_do_complete is not None:
            setattr(ipkernel.IPythonKernel, "do_complete", Hooker._orig_do_complete)
            Hooker._orig_do_complete = None

        if Hooker._orig_deduplicate is not None:
            setattr(Hooker._putils, "_deduplicate_completions", Hooker._orig_deduplicate)
            Hooker._orig_deduplicate = None



"""
Functions for working with numba.

Functionality includes:

- Functions for numba <-> riptable interop
- Convenience functions and decorators used when implementing numba-accelerated functions.

Notes
-----
Many functions in this module implement both a standard Python version and
an overload using `numba.extending.overload`. This is done to allow the code
to work when ``NUMBA_DISABLE_JIT=1`` is specified on the command line, e.g.
to allow functions to be debugged.
"""
__all__ = ["get_invalid", "get_max_valid", "get_min_valid", "is_valid"]

import numpy as np
import numpy.typing as npt
import numba as nb
from numba.extending import overload


def is_valid(x) -> bool:
    """
    A function for checking if data is valid.  This works for both floats and integers.

    - For floats, the invalid is NaN.
    - For signed integers, the invalid is the most NEGATIVE value of the type.
    - For unsigned integers, the invalid is the most POSITIVE value of the type.

    Parameters
    ----------
    x
        The value to check

    Returns
    -------
    A bool for whether the data is valid.
    """
    if x is None:
        return False

    x_dtype = x.dtype
    if np.issubdtype(x_dtype, np.floating):
        return not np.isnan(x)

    elif np.issubdtype(x_dtype, np.integer):
        dtype_iinfo = np.iinfo(x_dtype)
        inv_val = dtype_iinfo.min if np.issubdtype(x_dtype, np.signedinteger) else dtype_iinfo.max
        return x != inv_val

    elif np.issubdtype(x_dtype, bool) or np.issubdtype(x_dtype, np.bytes_) or np.issubdtype(x_dtype, str):
        return True

    else:
        raise TypeError(f"Valid detection has not been implemented for the type '{type(x)}'.")


@overload(is_valid)
def _is_valid_nb(x):
    """
    A function for checking if data is valid.  This works for both floats and integers.

    - For floats, the invalid is NaN.
    - For signed integers, the invalid is the most NEGATIVE value of the type.
    - For unsigned integers, the invalid is the most POSITIVE value of the type.

    Parameters
    ----------
    x
        The value to check

    Returns
    -------
    A bool for whether the data is valid.
    """
    if isinstance(x, nb.types.Float):
        return lambda x: not np.isnan(x)

    elif isinstance(x, nb.types.Integer):
        if x.signed:
            inv_val = x.minval
        else:
            inv_val = x.maxval
        return lambda x: x != inv_val

    elif isinstance(x, nb.types.NoneType):
        return lambda x: False

    else:
        raise TypeError("Valid detection has not been implemented for this type.")


def get_invalid(x):
    """
    A function for getting the invalid for a type of element.

    This works for both floats and integers.

    - For floats, the invalid is NaN.
    - For signed integers, the invalid is the most NEGATIVE value of the type.
    - For unsigned integers, the invalid is the most POSITIVE value of the type.

    For arrays, the invalid of the dtype of the array is returned.

    Parameters
    ----------
    x
        An element of the type you want the invalid for

    Returns
    -------
    The invalid value for `x`'s type/dtype.
    """
    x_dtype = x.dtype

    if np.issubdtype(x_dtype, np.floating):
        return np.nan

    elif np.issubdtype(x_dtype, np.integer):
        dtype_iinfo = np.iinfo(x_dtype)
        return dtype_iinfo.min if np.issubdtype(x_dtype, np.signedinteger) else dtype_iinfo.max

    else:
        raise TypeError(f'No invalid has not been implemented for type "{x}".')


@overload(get_invalid)
def _get_invalid_nb(x):
    """
    A numba function for getting the invalid for a type of element.

    This works for both floats and integers.

    - For floats, the invalid is NaN.
    - For signed integers, the invalid is the most NEGATIVE value of the type.
    - For unsigned integers, the invalid is the most POSITIVE value of the type.

    For arrays, the invalid of the dtype of the array is returned.

    Parameters
    ----------
    x
        An element of the type you want the invalid for

    Returns
    -------
    The invalid value
    """
    # If an array, get it's dtype
    if isinstance(x, nb.types.Array):
        x = x.dtype

    if isinstance(x, nb.types.Float):
        return lambda x: np.nan

    elif isinstance(x, nb.types.Integer):
        if x.signed:
            ret_val = x.minval
        else:
            ret_val = x.maxval
        return lambda x: ret_val

    elif isinstance(x, nb.types.NoneType):
        ret_val = nb.int64.minval
        return lambda x: ret_val

    else:
        raise TypeError(f'No invalid has not been implemented for type "{x}".')


def get_max_valid(x):
    """
    Get the maximum valid value for the dtype of an array or scalar.

    This function supports integer and floating-point values.

    Parameters
    ----------
    x
        An array or scalar value of the type you want the maximum valid value for.

    Returns
    -------
    max_valid
        The maximum valid value for `x`'s dtype.
    """
    x_dtype = x.dtype

    if np.issubdtype(x_dtype, np.floating):
        fi = np.finfo(x_dtype)
        if fi.bits == 64:
            return np.finfo(np.float64).max
        elif fi.bits == 32:
            return np.finfo(np.float32).max
        elif fi.bits == 16:
            return np.finfo(np.float16).max

    elif np.issubdtype(x_dtype, np.integer):
        dtype_iinfo = np.iinfo(x_dtype)
        return dtype_iinfo.max if np.issubdtype(x_dtype, np.signedinteger) else dtype_iinfo.max - 1

    elif np.issubdtype(x_dtype, bool):
        return True

    else:
        raise TypeError(f'get_max_valid() has not been implemented for type "{x}".')


@overload(get_max_valid)
def _get_max_valid_nb(x):
    """
    Get the maximum valid value for the dtype of an array or scalar.

    This function supports integer and floating-point values.

    Parameters
    ----------
    x
        An array or scalar value of the type you want the maximum valid value for.

    Returns
    -------
    max_valid
        The maximum valid value for `x`'s dtype.
    """
    # If an array, get it's dtype
    if isinstance(x, nb.types.Array):
        x = x.dtype

    if isinstance(x, nb.types.Float):
        if x.bitwidth == 64:
            ret_val = np.finfo(np.float64).max
        elif x.bitwidth == 32:
            ret_val = np.finfo(np.float32).max
        elif x.bitwidth == 16:
            ret_val = np.finfo(np.float16).max
        return lambda x: ret_val

    elif isinstance(x, nb.types.Integer):
        ret_val = x.maxval if x.signed else x.maxval - 1
        return lambda x: ret_val

    elif isinstance(x, nb.types.Boolean):
        return lambda x: True

    else:
        raise TypeError(f'_get_max_valid_nb() has not been implemented for type "{x}".')


def get_min_valid(x):
    """
    Get the minimum valid value for the dtype of an array or scalar.

    Parameters
    ----------
    x
        An array or scalar value of the type you want the minimum valid value for.

    Returns
    -------
    min_valid
        The minimum valid value for `x`'s dtype.
    """
    x_dtype = x.dtype

    if np.issubdtype(x_dtype, np.floating):
        fi = np.finfo(x_dtype)
        if fi.bits == 64:
            return np.finfo(np.float64).min
        elif fi.bits == 32:
            return np.finfo(np.float32).min
        elif fi.bits == 16:
            return np.finfo(np.float16).min

    elif np.issubdtype(x_dtype, np.integer):
        dtype_iinfo = np.iinfo(x_dtype)
        return dtype_iinfo.min + 1 if np.issubdtype(x_dtype, np.signedinteger) else dtype_iinfo.max

    elif np.issubdtype(x_dtype, bool):
        return False

    elif np.issubdtype(x_dtype, np.bytes_):
        return b""

    elif np.issubdtype(x_dtype, str):
        return ""

    else:
        raise TypeError(f'get_min_valid() has not been implemented for type "{x}".')


@overload(get_min_valid)
def _get_min_valid_nb(x):
    """
    Get the minimum valid value for the dtype of an array or scalar.

    Parameters
    ----------
    x
        An array or scalar value of the type you want the minimum valid value for.

    Returns
    -------
    min_valid
        The minimum valid value for `x`'s dtype.
    """
    # If an array, get it's dtype
    if isinstance(x, nb.types.Array):
        x = x.dtype

    if isinstance(x, nb.types.Float):
        if x.bitwidth == 64:
            ret_val = np.finfo(np.float64).min
        elif x.bitwidth == 32:
            ret_val = np.finfo(np.float32).min
        elif x.bitwidth == 16:
            ret_val = np.finfo(np.float16).min
        return lambda x: ret_val

    elif isinstance(x, nb.types.Integer):
        ret_val = x.minval + 1 if x.signed else x.minval
        return lambda x: ret_val

    elif isinstance(x, nb.types.Boolean):
        return lambda x: False

    elif isinstance(x, nb.types.Bytes):
        return lambda x: b""

    elif isinstance(x, nb.types.UnicodeType):
        return lambda x: ""

    else:
        raise TypeError(f'_get_min_valid_nb() has not been implemented for type "{x}".')



__all__ = ["SharedMemory"]

import os
import sys
import warnings

from .rt_enum import TypeRegister


class SharedMemoryMeta(type):
    SM_FOLDER = "/dev/shm/"
    GLOBAL_NAME = "Global_"
    SM_LIMIT = 2.5e10

    # ------------------------------------------------------------------------------------
    def listdir(cls, sharename=None):
        """
        (Linux only)
        returns list of all riptable shared memory files, or those with a specific sharename.
        """
        if sys.platform != "win32":
            prefix = cls.GLOBAL_NAME
            if sharename is not None:
                prefix += sharename + "!"
            plen = len(prefix)
            try:
                dir = os.listdir(cls.SM_FOLDER)
                return [f[plen:] for f in dir if f.startswith(prefix)]
            except Exception as e:
                return []

    # ------------------------------------------------------------------------------------
    def clear(cls, sharename=None):
        """
        (Linux only)
        :param sharename: (optional) specify a specific sharename

        Will remove all shared memory in /dev/shm, or all shared memory in the same directory with a specific share name.
        """
        if sys.platform != "win32":
            dir = cls._get_full_paths(sharename)
            for f in dir:
                os.remove(f)
        else:
            raise NotImplementedError(
                f"This feature is for Linux only. To clear shared memory in Windows, exit  your python session."
            )

    # ------------------------------------------------------------------------------------
    def view(cls, sharename=""):
        """
        (Linux only)
        TODO: add option for file info
        """
        return cls.listdir(sharename)

    # ------------------------------------------------------------------------------------
    def __repr__(cls):
        if sys.platform != "win32":
            dir = cls.listdir()
            if len(dir) == 0:
                return "Shared memory is empty."

            dir = "\n".join([f"Shared memory {cls.size}:"] + dir)
            return dir
        else:
            return "Shared memory functions not currently supported on Windows."

    # ------------------------------------------------------------------------------------
    def _get_full_paths(cls, sharename=None):
        """
        Get full file paths on linux shared memory. Necessary for deleting, checking file sizes.
        """
        fullpaths = []
        try:
            dir = os.listdir(cls.SM_FOLDER)
            prefix = cls.GLOBAL_NAME
            if sharename is not None:
                prefix += sharename + "!"
            for f in dir:
                if f.startswith(prefix):
                    fullpaths.append(cls.SM_FOLDER + f)
        except Exception as e:
            pass
        return fullpaths

    # ------------------------------------------------------------------------------------
    def _convert_bytes(cls, num):
        """
        this function will convert bytes to MB.... GB... etc
        """
        for x in ["bytes", "KB", "MB", "GB", "TB"]:
            if num < 1024.0:
                return "%3.1f %s" % (num, x)
            num /= 1024.0

    # ------------------------------------------------------------------------------------
    def _get_file_size(cls):
        dir = cls._get_full_paths()
        total_size = 0
        try:
            for f in dir:
                file_info = os.stat(f)
                total_size += file_info.st_size
        except Exception as e:
            pass
        return total_size

    # ------------------------------------------------------------------------------------
    def check_shared_memory_limit(cls):
        if sys.platform != "win32":
            size = cls._get_file_size()

            if size > cls.SM_LIMIT:
                readable = cls._convert_bytes(size)
                warnings.warn(f"!!!Shared memory is using {readable}. Consider using sm.clear() to remove.")

    # ------------------------------------------------------------------------------------
    @property
    def size(cls):
        s = cls._get_file_size()
        return cls._convert_bytes(s)


class SharedMemory(metaclass=SharedMemoryMeta):
    pass


TypeRegister.SharedMemory = SharedMemory



"""Unittests for riptable module level functions such as any, all, average, percentile, quantile, etc."""
import itertools

import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_array_equal

import riptable as rt
from riptable.Utils.teamcity_helper import is_running_in_teamcity


class TestCat2Keys:
    @pytest.mark.parametrize(
        "key1,key2",
        [
            (list("abc"), list("xyz")),
            (np.array(list("abc")), np.array(list("xyz"))),
            (np.array(list("abc")), list("xyz")),
            # add the following test cases for nested list of arrays and categoricals
            # ([FA(list('abc')), FA(list('def'))], [FA(list('uvw')), FA(list('xyz'))])
            # ([FA('a'), FA('b'), FA('c')], [FA('x'), FA('y'), FA('z')]),
            # (rt.Cat(list('abc'), ordered=True), rt.Cat(list('xyz'), ordered=True)),
        ],
    )
    def test_cat2keys(self, key1, key2):
        multi_cat = rt.cat2keys(key1, key2)
        assert len(key1) == len(key2)  # add test to check different length lists

        # these are the expected entries in the multi key categorical dictionary
        n = len(key1)
        expected_key1 = rt.FastArray([k for _ in range(n) for k in key1])
        expected_key2 = rt.FastArray([k for k in key2 for _ in range(n)])

        key_itr = iter(multi_cat.category_dict)
        actual_key1 = multi_cat.category_dict[next(key_itr)]
        actual_key2 = multi_cat.category_dict[next(key_itr)]

        assert_array_equal(expected_key1, actual_key1)
        assert_array_equal(expected_key2, actual_key2)

        # Taking the entries one by one of expected_key1 and expected_key2 should produce the
        # cartesian product of key1 and key2.
        expected_product = [k1 + k2 for k1, k2 in itertools.product(key1, key2)]
        actual_product = np.array(sorted([k1 + k2 for k1, k2 in zip(actual_key1, actual_key2)]), dtype="U2")
        assert_array_equal(sorted(expected_product), actual_product)


class TestQuantile:
    def test_unsupported_half_float_dtype(self):
        # Related Jira issue: RIP-341
        arr = np.zeros(shape=(2,), dtype=np.float16)
        with pytest.raises(SystemError):
            rt.quantile(arr, 0, "lower")

    def test_one_darray(self):
        arr = np.zeros(shape=(2,), dtype=np.float32)
        result, _, _ = rt.quantile(arr, 0, "lower")
        assert result == 0.0

    @pytest.mark.xfail(reason="RIP-341: supports shapes (x,), but not (x,0)")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    def test_one_darray_xfail(self):
        arr = np.zeros(shape=(2, 0), dtype=np.float32)
        result, _, _ = rt.quantile(arr, 0, "lower")
        assert result == 0.0


class TestSearchSorted:
    def _mask(self, arr):
        return arr % 2 == 0

    def test_simple(self):
        sides = {"left", "right"}

        arr = np.arange(10)
        v = arr[self._mask(arr)]

        for side in sides:
            rt_indicies = rt.searchsorted(arr, v, side)
            np_indicies = np.searchsorted(arr, v, side)
            assert_array_equal(
                rt_indicies,
                np_indicies,
                err_msg=f"array {arr} values to insert {v} side {side}",
            )

    def test_simple_left(self):
        side = "left"

        arr = np.array([0, 0, 0, 1])
        v = arr[self._mask(arr)]

        rt_indicies = rt.searchsorted(arr, v, side)
        np_indicies = np.searchsorted(arr, v, side)
        assert_array_equal(
            rt_indicies,
            np_indicies,
            err_msg=f"using array {arr}\nvalues to insert {v}\nside {side}",
        )

    @pytest.mark.xfail(reason="RIP-345 - expected insertion points at position 3, got 2")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    def test_simple_right_xfail(self):
        side = "right"

        arr = np.array([0, 0, 0, 1])
        v = arr[self._mask(arr)]

        rt_indicies = rt.searchsorted(arr, v, side)
        np_indicies = np.searchsorted(arr, v, side)
        assert_array_equal(
            rt_indicies,
            np_indicies,
            err_msg=f"using array {arr}\nvalues to insert {v}\nside {side}",
        )

    def test_simple_right2(self):
        side = "right"

        arr = np.array([0, 2, 2, 2, 2])
        v = arr[self._mask(arr)]

        rt_indicies = rt.searchsorted(arr, v, side)
        np_indicies = np.searchsorted(arr, v, side)
        assert_array_equal(
            rt_indicies,
            np_indicies,
            err_msg=f"using array {arr}\nvalues to insert {v}\nside {side}",
        )

    @pytest.mark.xfail(reason="RIP-345: expected insertion points at position 1, got 2")
    @pytest.mark.skipif(is_running_in_teamcity(), reason="Please remove alongside xfail removal.")
    def test_simple_left2_xfail(self):
        side = "left"

        arr = np.array([0, 2, 2, 2, 2])
        v = arr[self._mask(arr)]

        rt_indicies = rt.searchsorted(arr, v, side)
        np_indicies = np.searchsorted(arr, v, side)
        assert_array_equal(
            rt_indicies,
            np_indicies,
            err_msg=f"using array {arr}\nvalues to insert {v}\nside {side}",
        )

    def test_searchsorted(self):
        a = rt.arange(10.0)
        b = rt.arange(20.0) / 2
        b[3] = -np.inf
        b[7] = np.inf
        b[5] = np.nan
        b[2] = 100
        b[1] = -100
        x1 = np.searchsorted(a, b, side="left")
        x2 = rt.searchsorted(a, b, side="left")
        assert sum(x1 - x2) == 0

        x1 = rt.searchsorted(a, b, side="right")
        x2 = rt.searchsorted(a, b, side="right")
        assert sum(x1 - x2) == 0

        b = b.astype(np.int32)


class TestStd:
    @pytest.mark.parametrize(
        "a",
        # ACTUAL:  0.5345224838248488
        # DESIRED: 0.0
        [
            [
                1801439850948199,
                1801439850948199,
                1801439850948199,
                1801439850948199,
                1801439850948199,
                1801439850948199,
                1801439850948199,
                1801439850948199,
            ]
        ],
    )
    def test_std(self, a):
        arr = np.array(a, dtype=np.int64)
        np_std = np.std(arr, ddof=1)
        rt_std = rt.std(rt.FastArray(arr))
        assert_almost_equal(rt_std, np_std, decimal=6)


class TestSum:
    # add tests around promotion to larger type
    # e.g. sum of int32 type that overflows will be a result of int64 type
    # this behavior differs from numpy.sum
    @pytest.mark.xfail(reason="RIP-351 - sum not working as expected for some inputs")
    @pytest.mark.parametrize(
        "a",
        # ACTUAL:  3.193344496000001e+292
        # DESIRED: 3.193344496e+292
        [
            [
                3.99168062e291,
                3.99168062e291,
                3.99168062e291,
                3.99168062e291,
                3.99168062e291,
                3.99168062e291,
                3.99168062e291,
                3.99168062e291,
            ]
        ],
    )
    def test_sum_float(self, a):
        arr = np.array(a, dtype=np.float64)
        np_sum = np.sum(arr)
        rt_sum = rt.sum(rt.FastArray(arr))
        assert_almost_equal(rt_sum, np_sum, decimal=3)


class TestWhere:
    def test_where(self):
        arr = np.array([0], dtype=np.int64)
        min = arr.min()
        np_gt_min = np.where(arr > min)
        rt_gt_min = rt.where(arr > min)
        assert_array_equal(
            np_gt_min,
            rt_gt_min,
            err_msg=f"array elements greater than the minimum {min}",
        )

    @pytest.mark.xfail(reason="expected dtype of the original type")
    def test_where_dtype_demoting(self):
        dtype = np.int64
        arr = np.array([0, 1], dtype=dtype)
        min, mean, max = arr.min(), arr.mean(), arr.max()
        np_gt_min = np.where(arr > min)
        rt_gt_min = rt.where(arr > min)
        assert_array_equal(
            np_gt_min,
            rt_gt_min,
            err_msg=f"array elements greater than the minimum {min}",
        )
        assert dtype == rt_gt_min[0].dtype, "expected dtype of the original type"



__all__ = [
    "IMatrix",
]

from numpy import array, empty

from .rt_enum import TypeRegister
from .rt_numpy import vstack


class IMatrix:
    """
    Experimental class designed to take a Dataset and make a 2d matrix efficiently.
    It uses rt.vstack order='F' which uses rt.hstack plus np.reshape.

    The matrix is shaped so that it can be inserted back into the Dataset.

    Other Parameters
    ----------------
    dtype:
    order:
    colnames:

    See Also
    --------
    rt.vstack

    """

    def __init__(self, ds, dtype=None, order="F", colnames=None):
        if not isinstance(ds, TypeRegister.Dataset):
            raise TypeError(f"The first argument must be a dataset")

        if not isinstance(colnames, list):
            raise TypeError(f"Pass in a list of column names such as ['Exch1','Exch2', 'Exch3']")

        self.rebuild(ds, dtype=dtype, order=order, colnames=colnames)

    @property
    def dataset(self):
        return self._dataset

    @property
    def imatrix(self):
        return self._imatrix

    def rebuild(self, ds=None, dtype=None, order="F", colnames=None):
        if ds is None:
            ds = self._dataset

        if colnames is None:
            colnames = [*ds.keys()]

        # make the matrix from the dataset
        self._imatrix = vstack([*ds[colnames].values()], dtype=dtype, order=order)

        # make a dataset from the matrix
        self._dataset = TypeRegister.Dataset({c: self._imatrix[:, i] for i, c in enumerate(colnames)})

    def __getitem__(self, fld):
        """
        row slicing
        """
        # apply the row mask or slice
        self._dataset = self._dataset[fld]

        # have to reconstruct the 2d array since it has been sliced
        self.rebuild()

    # -------------------------------------------------------
    # 2d arithmetic functions.
    def apply2d(self, func, name=None, showfilter=True):
        """
        Parameters
        ----------
        func: function or method name of function

        Returns
        -------
        X and Y axis calculations

        """
        imatrix = self._imatrix

        if not callable(func):
            func = getattr(array, func)

        if callable(func):
            if name is None:
                name = func.__name__
                name = str.capitalize(name)

            row_count, col_count = imatrix.shape

            # horizontal func
            # print("im0", imatrix.nansum())
            totalsY = func(imatrix, axis=1)

            # possibly remove filtered top row
            if not showfilter:
                totalsY = totalsY[1:]

            # reserve an extra for the total of totals
            totalsX = empty(col_count + 1, dtype=totalsY.dtype)

            # consider #imatrix.nansum(axis=0, out=totalsX)
            for i in range(col_count):
                arrslice = imatrix[:, i]

                # possibly skip over first value
                if not showfilter:
                    arrslice = arrslice[1:]

                totalsX[i] = func(arrslice)

            # calc total of totals - cell on far right and bottom
            totalsX[-1] = func(totalsY)

            return totalsX, totalsY

        return None, None



__all__ = [
    "class_error",
    "groupScatter",
    "linear_spline",
    "lm",
    "mae",
    "plotPrediction",
    "plot_hist",
    "r2",
    "statx",
    "winsorize",
]

import numpy as np

# extra classes
import pandas as pd

import riptable as rt

from .rt_enum import TypeRegister
from .rt_numpy import zeros

# bokeh is an optional dependency.
try:
    from bokeh.plotting import figure, output_notebook, show

    # from bokeh.models import Label
except ImportError:
    pass


# TODO: Organize the functions in here better
# TODO: Add documentation
# TODO: Replace pandas dependence with display util
def statx(X):
    if not isinstance(X, np.ndarray):
        X = np.array(X)
    pVals = [0.1, 1, 10, 25, 50, 75, 90, 99, 99.9]
    pValNames = [
        "min",
        "0.1%",
        "1%",
        "10%",
        "25%",
        "50%",
        "75%",
        "90%",
        "99%",
        "99.9%",
        "max",
        "Mean",
        "StdDev",
        "Count",
        "NaN_Count",
    ]
    filt = np.isfinite(X)
    X_sub = X[filt]
    if len(X_sub) > 0:
        vals = np.percentile(X_sub, pVals)
        vals = np.insert(vals, 0, np.min(X_sub))
        vals = np.append(vals, np.max(X_sub))
        vals = np.append(vals, np.mean(X_sub))
        vals = np.append(vals, np.std(X_sub))
    else:
        vals = np.array(13 * [rt.nan])
    validcount = np.sum(filt)

    # plain count
    vals = np.append(vals, X.size)

    # nancount
    vals = np.append(vals, np.sum(np.isnan(X)))
    out = pd.DataFrame({"Stat": pValNames, "Value": vals})
    return out


# NOTE: people might prefer name clip/bound?
def winsorize(Y, lb, ub):
    out = np.maximum(np.minimum(Y, ub), lb)
    return out


def plot_hist(Y, bins):
    df = pd.DataFrame({"Y": Y})
    df.hist(bins=bins)


def r2(X, Y):
    # why are these flipped back?
    xmean = np.mean(X)
    ymean = np.mean(Y)
    xycov = np.mean(np.multiply((X - xmean), (Y - ymean)))
    xvar = np.var(X)
    yvar = np.var(Y)
    r2_value = (xycov * xycov) / (xvar * yvar)
    return r2_value


def mae(X, Y):
    return np.nanmean(np.abs(X - Y))


def class_error(X, Y):
    X2 = np.round(X)
    Y2 = np.round(Y)
    class_err = np.sum(np.abs(X2 - Y2)) / X2.shape[0]
    return class_err


def lm(X, Y, intercept=True, removeNaN=True, displayStats=True):
    # TODO: Better display for stats
    X0 = X.copy()
    Y0 = Y.copy()
    if len(X0.shape) == 1:
        X0 = X0.reshape(X0.shape[0], 1)
    if len(Y0.shape) == 1:
        Y0 = Y0.reshape(Y0.shape[0], 1)
    if intercept:
        X0 = np.hstack([np.ones((X0.shape[0], 1)), X0])

    if removeNaN:
        goodData = ~np.isnan(np.sum(X0, axis=1)) & ~np.isnan(np.sum(Y0, axis=1))
        X0 = X0[goodData, :]
        Y0 = Y0[goodData, :]
    VXX = np.matmul(np.transpose(X0), X0)
    VXY = np.matmul(np.transpose(X0), Y0)
    coeff = np.linalg.solve(VXX, VXY)

    if displayStats:
        YHat = np.matmul(X0, coeff)
        err = Y0 - YHat
        err = err.reshape(Y0.shape[0], Y0.shape[1])
        RMS = np.sqrt(np.mean(err * err))
        MAE = np.mean(np.abs(err))
        A = np.linalg.solve(VXX, np.transpose(X0))
        SE = np.sqrt(np.sum(A * A, axis=1)) * RMS
        tStat = coeff / SE.reshape(coeff.shape[0], 1)
        R = np.mean(YHat * Y0) / (np.std(YHat) * np.std(Y0))
        R2 = R * R

        print("R2 = ", R2)
        print("RMSE = ", RMS)
        print("MAE = ", MAE)
        print("tStats: ")
        print(tStat)
    return coeff


def linear_spline(X0, Y0, knots, display=True):
    X = X0.copy()
    Y = Y0.copy()
    X = X.reshape(X.shape[0], 1)
    Y = Y.reshape(Y.shape[0], 1)
    knots.sort()
    numKnots = len(knots)
    goodData = ~np.isnan(X) & ~np.isnan(Y)
    X = X[goodData]
    Y = Y[goodData]
    XAug = np.nan * np.zeros((X.shape[0], 2 + numKnots))
    XAug[:, 0] = np.ones_like(X)
    XAug[:, 1] = X

    for j in range(numKnots):
        XAug[:, 2 + j] = np.maximum(X - knots[j], 0.0)

    coeff = lm(XAug, Y, intercept=False, removeNaN=True, displayStats=False)
    YHat = np.matmul(XAug, coeff)

    X_uniq, X_idx = np.unique(X, return_index=True)
    YHat_uniq = YHat[X_idx]
    # idx = X_uniq <= ub & X_uniq >= lb

    output_notebook()
    # create a new plot
    p = figure(tools="pan,box_zoom,reset,save", title="example", x_axis_label="sections", y_axis_label="particles")

    p.circle(X_uniq.flatten(), YHat_uniq.flatten(), legend="y", fill_color="red", size=8)
    if display:
        show(p)
    return knots, coeff


# TODO: Make formatting aware of environment, e.g., Spyder, jupyter, etc. in groupScatter and plotPrediction
# NOTE: Can we use regPlot from seaborn
# won't display in jupyter lab
# better auto-detect bounds
# suppress nan warnings


def plotPrediction(X, Yhat, Y, N, lb=None, ub=None):

    if lb is None:
        lowerBound = np.nanmin(X)
    else:
        lowerBound = lb

    if lb is None:
        upperBound = np.nanmax(X)
    else:
        upperBound = ub

    goodFilt = (
        np.isfinite(X) & np.isfinite(Y) & (X <= upperBound) & (X >= lowerBound) & np.isfinite(Yhat) & np.isfinite(Y)
    )

    dF = pd.DataFrame({"X": X[goodFilt], "Y": Y[goodFilt], "Yhat": Yhat[goodFilt]})
    dF.sort_values("X", inplace=True)
    dF.reset_index(drop=True, inplace=True)
    groupSize = np.floor(dF.shape[0] / N)
    dF["Bucket"] = np.int32(np.floor(dF.index.values / groupSize))
    out = dF.groupby("Bucket")[["X", "Y", "Yhat"]].agg(np.nanmean)

    output_notebook()
    # output_file("test.html")
    p = figure(tools="pan,box_zoom,reset,save", title="example", x_axis_label="sections", y_axis_label="particles")

    p.circle(out.X, out.Y, legend="y", fill_color="red", size=8)
    p.circle(out.X, out.Yhat, legend="yhat", fill_color="blue", size=8)
    show(p)

    # xloc = np.min(out.X)
    # yloc = np.max(out.Y)
    # plt.text(xloc,yloc,r'$r^2 =$ {0:.2e}'.format(r2Val),fontsize=16)


def polyFit(x, y, d=1, filter=None):
    """
    Fits a polynimial least-squares regression.

    Parameters
    ----------
    x : ndarray
        The x data points.
    y : ndarray
        The y data points.
    d : integer, optional
        The degree of regression (d=1 for standard linear regression).
    filter : ndarray, optional
        Boolean mask of data to include

    Returns
    -------
    c : ndarray
        The array of coefficients of the polynomial (constant term first).

    """
    if filter is not None:
        x = x[filter]
        y = y[filter]
    A = np.vander(x, d + 1, increasing=True)
    AtA = np.matmul(A.transpose(), A)
    Ay = np.matmul(A.transpose(), y)
    c = np.linalg.solve(AtA, Ay)
    return c


def groupScatter(*arg, **kwarg):
    """
    This function has been moved to playa.stats.
    """
    raise NotImplementedError("This function has been moved to playa.plot")



# -*- coding: utf-8 -*-
"""
Benchmarks for 'merge' functions.
"""
__all__ = [
    "create_trial_dataset",
    "elapsed_ns_colname",
    "implementation_colname",
    "loop_iter_colname",
    "quick_analysis",
]


import inspect
from collections import namedtuple
from contextlib import contextmanager
from functools import wraps
from itertools import product
from typing import Callable, List, Mapping, Optional, Union

import numpy as np
import riptide_cpp as rc

from ..rt_categorical import Categorical
from ..rt_dataset import Dataset
from ..rt_enum import INVALID_SHORT_NAME
from ..rt_fastarray import FastArray
from ..rt_multiset import Multiset
from ..rt_numpy import arange, full
from .benchmark import timestamper

elapsed_ns_colname: str = "elapsed_ns"
"""
Name to use in benchmark Datasets for the column holding the elapsed time (in ns)
for the function invocations. 
"""

loop_iter_colname: str = "loop_idx"
"""
Name to use in benchmark Datasets for the column holding the loop index
for each function invocation. 
"""

implementation_colname: str = "impl"
"""
Name to use in benchmark _comparison_ Datasets for the column holding the
implementation name.
"""

# TODO move this into a settings library file.
def set_thread(thread_count):
    # Consdier setting wake up threads to the number of worker threads, i.e.
    # thread_count minus the main thread.
    rc.SetThreadWakeUp(thread_count)


def set_recycler(recycle):
    if recycle:
        FastArray._RON(quiet=True)
    else:
        FastArray._ROFF(quiet=True)


def get_thread():
    thread_count = rc.GetThreadWakeUp()
    return thread_count


def get_recycler():
    return FastArray.Recycle


SettingEntry = namedtuple("SettingEntry", ("getter_func", "setter_func"))
"""SettingEntry consists of a pointer to a setting getter and setter function.
The call to `setter_func(getter_func())` should set the original state.
"""

BENCHMARK_SETTINGS_TABLE = {
    # TODO - FasterUfunc, Recycler timeouts, and MathLedger settings
    "thread_count": SettingEntry(get_thread, set_thread),
    "recycler": SettingEntry(get_recycler, set_recycler),
}
"""BENCHMARK_SETTINGS_TABLE is the global map of setting names to SettingEntries
that allow benchmarks to apply stateful settings. E.g., setting riptable thread
counts or recycler modes that vary across benchmark runs.
"""


def get_settings_getter_func(key: str) -> Optional[Callable]:
    if key in BENCHMARK_SETTINGS_TABLE:
        return BENCHMARK_SETTINGS_TABLE[key].getter_func
    return None


def get_settings_setter_func(key: str) -> Optional[Callable]:
    if key in BENCHMARK_SETTINGS_TABLE:
        return BENCHMARK_SETTINGS_TABLE[key].setter_func
    return None


def create_comparison_dataset(bench_datasets: Mapping[str, Dataset], destroy: bool = False) -> Dataset:
    """
    TODO

    Parameters
    ----------
    bench_datasets
    destroy : bool, defaults to False

    Returns
    -------
    compare_dataset : Dataset
    """
    # For each dataset, create a shallow copy (to avoid modifying the original),
    # then add a Categorical column with the implementation name to it.
    impl_datasets: List[Dataset] = []
    for impl_name, impl_bench_data in bench_datasets.items():
        data_copy = impl_bench_data if destroy else impl_bench_data.copy(deep=False)
        data_copy[implementation_colname] = Categorical([impl_name]).tile(len(data_copy))
        data_copy.col_move_to_front([implementation_colname])
        impl_datasets.append(data_copy)

    # hstack the Datasets together and return.
    return Dataset.hstack(impl_datasets, destroy=destroy)


def create_trial_dataset(
    timing_data: np.ndarray,
    trial_params: Optional[Mapping[str, Union[int, str, bytes, float, np.dtype]]],
) -> Dataset:
    """
    Create the Dataset instance for a trial (repeated invocations of a function to be benchmarked).

    Parameters
    ----------
    timing_data : FastArray
        Array containing the elapsed time, in nanoseconds, for each invocation
        of the function being benchmarked.
    trial_params : dict
        A dictionary containing the trial parameters by name.
        Parameter names should all be valid for use as column names in a Dataset.

    Returns
    -------
    trial_dataset : Dataset
    """
    trial_data = Dataset()

    # Add the timing data and loop index columns.
    trial_length = len(timing_data)
    trial_data[loop_iter_colname] = arange(trial_length)
    trial_data[elapsed_ns_colname] = timing_data

    # If trial parameters were specified, add them to the dataset.
    if trial_params is not None:
        for param_name, param_value in trial_params.items():
            # Create an array for this parameter the same length as the number of function invocations,
            # and fill it with the parameter value. For string-valued parameters, create a Categorical
            # so the array is more compact in memory.
            if isinstance(param_value, int):
                dtype = np.min_scalar_type(param_value)
                param_val_array = full(trial_length, param_value, dtype)
            elif isinstance(param_value, (bytes, str)):
                dtype = np.min_scalar_type(param_value)
                param_val_array = Categorical(full(trial_length, param_value, dtype))
            elif isinstance(param_value, float):
                param_val_array = full(trial_length, param_value, dtype=np.float64)
            elif isinstance(param_value, np.dtype):
                # Same as for the bytes/str case; we just extract the dtype name and use that.
                param_value = param_value.name
                dtype = np.min_scalar_type(param_value)
                param_val_array = Categorical(full(trial_length, param_value, dtype))
            else:
                raise ValueError(
                    f"Unsupported type '{type(param_value)}' specified for trial parameter '{param_name}'."
                )

            # Add the trial parameter array to the Dataset.
            trial_data[param_name] = param_val_array

    # Return the constructed Dataset.
    return trial_data


def quick_analysis(bench_data: Dataset) -> Multiset:
    """
    Performs a simple, quick analysis on raw benchmark data

    Parameters
    ----------
    bench_data : Dataset
        A Dataset created with `create_trial_dataset`, or a Dataset
        created by `hstack`-ing such Datasets (from multiple trials) together.

    Returns
    -------
    quick_bench_analysis : Multiset
    """
    # Group the results by all columns _except_ the per-invocation columns (the loop index and elapsed time);
    # this effectively gets us a group per trial.
    gb_keys = list(bench_data.keys())
    gb_keys.remove(loop_iter_colname)
    gb_keys.remove(elapsed_ns_colname)
    gb_trial = bench_data.groupby(gb_keys)

    # Get the min/median/max elapsed time per trial and return it as a Multiset for display.
    return gb_trial[elapsed_ns_colname].agg(["min", "median", "max"])


@contextmanager
def _benchmark_setting_context(setting_to_values):
    # getter / setter for benchmark setting context
    # check all settings have defined getters and setters and they're callable

    # save original values
    setting_to_originals = dict.fromkeys(setting_to_values)
    for k, v in setting_to_values.items():
        setting_to_originals[k] = get_settings_getter_func(k)()
    try:
        # set new values
        for k, v in setting_to_values.items():
            get_settings_setter_func(k)(v)
        yield
    finally:
        # set original values
        for k, v in setting_to_originals.items():
            get_settings_setter_func(k)(v)


def benchmark(*_benchmark_arguments, **_benchmark_kwargs):
    """main entry point for riptable benchmarking library"""

    def decorate(func):
        # validate function arguments
        if inspect.isclass(func):
            # @benchmark assumes it's decorating a function
            raise ValueError(f"'@benchmark' cannot be applied to a class {func.__name__}")

        # TODO handle benchmark functions that are instance or static methods
        # 'cls' and 'self' will appear in the argspec

        argspec_args = inspect.getfullargspec(func)[0]
        # TODO handle positional arguments from _benchmark_arguments
        benchmark_kwargs = dict(_benchmark_kwargs)

        # Handle required keyword arguments that are core to the benchmark framework.
        if "benchmark_params" not in benchmark_kwargs:
            raise ValueError(
                f"'benchmark_params' is a required keyword argument, got keyword arguments {benchmark_kwargs}"
            )
        benchmark_params = benchmark_kwargs["benchmark_params"]
        bench_iters = benchmark_kwargs.get("benchmark_iterations", 5)
        warmup_iters = _benchmark_kwargs.get("warmup_iterations", 1)

        # Benchmark parameter and setting checks.
        func_params = set(argspec_args)
        setting_params = benchmark_params.keys() - func_params
        param_symmetric_diff = func_params.union(setting_params).symmetric_difference(benchmark_params.keys())
        if param_symmetric_diff:
            raise ValueError(
                f"'@benchmark', when benchmarking function {func.__name__}, got param and / or settings that are not in benchmark_params.\nBenchmark parameters {benchmark_params}.\nBenchmark function parameters {func_params}\nBenchmark settings {setting_params}\nSymmetric difference between {param_symmetric_diff}"
            )

        @wraps(func)
        def wrapper(*args, **kwargs) -> Optional[Dataset]:
            benchmark_results: List[Dataset] = []
            # param_product contains a product sweep of all the parameters parameter values.
            param_product = product(*[benchmark_params[param_name] for param_name in benchmark_params.keys()])
            for i, param_values in enumerate(param_product):
                # Maintain two dictionaries for this trials parameters and settings values.
                setting_param_to_value = {}
                trial_param_to_value = dict.fromkeys(benchmark_params, INVALID_SHORT_NAME)
                for k, v in zip(benchmark_params.keys(), param_values):
                    if k in setting_params:
                        setting_param_to_value[k] = v
                    trial_param_to_value[k] = v
                with _benchmark_setting_context(setting_param_to_value):
                    timing_data = np.empty(bench_iters, dtype=np.int64)

                    # Copy trial parameters as arguments to func.
                    for arg_spec in argspec_args:
                        if arg_spec in trial_param_to_value:
                            kwargs[arg_spec] = trial_param_to_value[arg_spec]

                    # warm up
                    for _ in range(warmup_iters):
                        func(*args, **kwargs)

                    # trial run
                    for trial_idx in range(bench_iters):
                        start_time_ns = timestamper()
                        # Note, there is overhead when unpacking positional and keyword arguments within timestamper.
                        # Running cProfile with Python 3.7.1 on a dummy function that takes four positional
                        # and four keyword arguments showed the splat operator has some overhead
                        # - 100,000 iterations has a difference of 0.007 seconds
                        # - 1 million iterations has a difference of 0.071 seconds
                        # - 10 million iterations has a difference of 0.801 seconds
                        # more compared to normal parameter passing.
                        func(*args, **kwargs)
                        timing_data[trial_idx] = timestamper() - start_time_ns

                    # TODO - move this to its own function process_trial_params
                    # Should handle print friendly representations
                    for k, v in trial_param_to_value.items():
                        if isinstance(v, tuple):
                            trial_param_to_value[k] = str(v)
                        if isinstance(v, (np.ndarray, Dataset)):
                            trial_param_to_value[k] = len(v)

                    trial_data = create_trial_dataset(timing_data, trial_param_to_value)
                    benchmark_results.append(trial_data)

            # If benchmark_results is empty, there's nothing to be stacked;
            # emit a warning then return None. Attempting to stack an empty list
            # of Datasets results in a ValueError being raised, which may end up
            # causing the runner to crash and the results for the whole benchmark suite
            # not being saved.
            if benchmark_results:
                return Dataset.hstack(benchmark_results, destroy=True)
            else:
                return None

        return wrapper

    return decorate



import numpy as np
import pytest
from numpy.testing import assert_array_equal

import riptable as rt
from riptable.rt_utils import crc_match


@pytest.mark.parametrize(
    "arrs,expected",
    [
        # Some basic positive test cases.
        pytest.param(
            [
                rt.FastArray([b"ABCDEFGHIJKL"], dtype="|S32"),
                rt.FastArray([b"ABCDEFGHIJKL"], dtype="|S32"),
            ],
            True,
            id="S32__S32",
        ),
        # Test cases for verifying array shape is checked even when
        # calculated CRC value of the raw data is the same.
        pytest.param([np.arange(1, 10), np.arange(10)], False, id="int__int__leading-zero"),
        pytest.param(
            [
                rt.FastArray([b"2019/12/21"], dtype="|S32"),
                rt.FastArray([b"", b"2019/12/21"], dtype="|S32"),
            ],
            False,
            id="ascii__ascii__leading-empty",
        ),
    ],
)
def test_crc_match(arrs, expected):
    result = crc_match(arrs)
    assert result == expected


def test_mbget_no_default_uses_invalid():
    data = np.arange(start=3, stop=53, dtype=np.int8).view(rt.FA)
    indices = rt.FA([0, 25, -40, 17, 100, -80, 50, -51, 35])

    valid_indices = np.logical_and(indices >= -50, indices < 50)

    # Call the 'mbget' function without providing an explicit default value.
    result = rt.mbget(data, indices)

    # The resulting array should have the same dtype as the values/data array.
    assert data.dtype == result.dtype, "The result has a different dtype than the values/data array."

    # The elements with out-of-bounds indices should have been assigned the
    # riptable NA/sentinel value because a default value was not explicitly specified.
    assert_array_equal(valid_indices, rt.isnotnan(result))

    # Check that the valid indices fetched the correct values.
    assert_array_equal(rt.FA([3, 28, 13, 20, 38]), result[valid_indices])


def test_mbget_with_explicit_default():
    data = np.arange(start=3, stop=53, dtype=np.int8).view(rt.FA)
    indices = rt.FA([0, 25, -40, 17, 100, -80, 50, -51, 35])

    valid_indices = np.logical_and(indices >= -50, indices < 50)

    # Call the 'mbget' function, providing an explicit default value.
    default_value = 123
    result = rt.mbget(data, indices, d=default_value)

    # The resulting array should have the same dtype as the values/data array.
    assert data.dtype == result.dtype, "The result has a different dtype than the values/data array."

    # The elements with out-of-bounds indices should have been assigned the
    # explicitly-specified default value.
    assert_array_equal(valid_indices, result != default_value)

    # Check that the valid indices fetched the correct values.
    assert_array_equal(rt.FA([3, 28, 13, 20, 38]), result[valid_indices])


@pytest.mark.xfail(
    reason="BUG mbget does not widen the output dtype to accommodate the default value, nor does it validate the "
    "default value will fit into the data/values dtype, so the default value is silently truncated."
)
def test_mbget_with_too_large_explicit_default():
    data = np.arange(start=3, stop=53, dtype=np.int8).view(rt.FA)
    indices = rt.FA([0, 25, -40, 17, 100, -80, 50, -51, 35])

    valid_indices = np.logical_and(indices >= -50, indices < 50)

    # Call the 'mbget' function, providing an explicit default value
    # which is too large to be represented by the dtype of the data/values array.
    default_value = 1234
    result = rt.mbget(data, indices, d=default_value)

    # The resulting array will need to have a larger dtype than the original data array
    # to accommodate the explicit default value that was too large for the data's dtype.
    assert data.dtype != result.dtype, "The result has the same dtype as the values/data array."
    assert (
        np.dtype(data.dtype).itemsize < np.dtype(result.dtype).itemsize
    ), "The result dtype is not larger than the values/data dtype."

    # The elements with out-of-bounds indices should have been assigned the
    # explicitly-specified default value.
    assert_array_equal(valid_indices, result != default_value)

    # Check that the valid indices fetched the correct values.
    assert_array_equal(rt.FA([3, 28, 13, 20, 38]), result[valid_indices])


def test_mbget_with_non_FA_subclass():
    """Test how mbget works when the 'origin' array is NOT a FastArray (or a subclass of FA)."""
    data = np.arange(start=3, stop=53, dtype=np.int8)
    indices = rt.FA([0, 25, -40, 17, 100, -80, 50, -51, 35])

    valid_indices = np.logical_and(indices >= -50, indices < 50)

    # Call the 'mbget' function without providing an explicit default value.
    result = rt.mbget(data, indices)

    # The resulting array should have the same dtype as the values/data array.
    assert data.dtype == result.dtype, "The result has a different dtype than the values/data array."

    # The elements with out-of-bounds indices should have been assigned the
    # riptable NA/sentinel value because a default value was not explicitly specified.
    assert_array_equal(valid_indices, rt.isnotnan(result))

    # Check that the valid indices fetched the correct values.
    assert_array_equal(rt.FA([3, 28, 13, 20, 38]), result[valid_indices])
    assert type(data) == np.ndarray



import os
import unittest
from math import isclose

from riptable import *
from riptable import Categorical, Dataset, FastArray, GroupBy, arange, isnan, isnotnan
from riptable.rt_enum import INVALID_DICT
from riptable.tests.utils import LikertDecision
from riptable.Utils.pandas_utils import dataset_as_pandas_df

num_list = [3, 4, 5, 1, 2]
num_list_sorted = [1, 2, 3, 4, 5]
arr_types = [
    np.int8,
    np.uint8,
    np.int16,
    np.uint16,
    np.int32,
    np.uint32,
    np.int64,
    np.uint64,
    np.float32,
    np.float64,
]
arr_types_string = [np.bytes_, np.str_]
test_data = {"bool": np.array([True, False, True, False, True], dtype=bool)}
for dt in arr_types + arr_types_string:
    test_data[dt.__name__] = np.array(num_list, dtype=dt)
test_data["categorical"] = Categorical([str(i) for i in num_list])
all_headers = list(test_data.keys())
ds = Dataset(test_data)
gb_funcs = ["sum", "mean", "first", "last", "median", "min", "max", "var"]
gb_nan_funcs = ["nansum", "nanmean", "nanmedian", "nanvar"]  #'rolling', 'cumsum', 'nth'


class Groupby_Test(unittest.TestCase):
    def test_math_ops_same_return(self):
        result_dict = {
            "sum": [5, 10],
            "nansum": [5, 10],
            # TODO: add support for min / max on strings
            "min": [1, 2],
            "max": [4, 5],
        }
        gb = ds.gb("bool")
        for name, correct in result_dict.items():
            call = getattr(gb, name)
            ds_result = call()
            for dt in arr_types:
                column = getattr(ds_result, dt.__name__)
                is_correct = bool(np.all(column == np.array(correct, dtype=dt)))
                self.assertTrue(
                    is_correct,
                    msg=f"Incorrect result for keys {ds_result.label_get_names()} in column with datatype {dt} after {name} operation.",
                )

            # make sure string columns are not included
            for dt in arr_types_string:
                was_included = hasattr(ds_result, dt.__name__)
                self.assertFalse(
                    was_included,
                    msg=f"String-like column {dt.__name__} was included in {name} math operation.",
                )

    def test_filter(self):
        correct = [1, 2]
        result = ds.gb("bool", filter=[False, False, False, True, True]).min()
        for dt in arr_types:
            column = getattr(result, dt.__name__)
            is_correct = bool(np.all(column == np.array(correct, dtype=dt)))
            self.assertTrue(
                is_correct,
                msg=f"Incorrect result in column of datatype {dt} when filter was applied on sorted groupby. {column} {correct}",
            )

    def test_math_ops_float_return(self):
        gb = ds.gb("bool")
        result_dict = {
            "mean": [2.5, 3.333333333],
            "std": [2.121320343560, 1.527525231652],
            "median": [2.5, 3],
            "var": [4.5, 2.333333333333],
        }
        for name, correct in result_dict.items():
            call = getattr(gb, name)
            ds_result = call()
            for dt in arr_types:
                column = getattr(ds_result, dt.__name__)
                for idx, item in enumerate(column):
                    is_accurate = isclose(item, correct[idx], rel_tol=1e-06)
                    self.assertTrue(
                        is_accurate,
                        msg=f"Incorrect result for keys {ds_result.label_get_names()} in column with datatype {dt} after {name} operation. {item} != {correct[idx]}",
                    )

            # make sure string columns are not included
            for dt in arr_types_string:
                was_included = hasattr(ds_result, dt.__name__)
                self.assertFalse(
                    was_included,
                    msg=f"String-like column {dt.__name__} was included in {name} math operation.",
                )

    def test_keys_included(self):
        """
        By default, groupby keys will not be included in the dataset's main data for the groupby calculation.
        The keyword include_keys can be set to include them.
        """
        # now included
        ds_result = ds.gb("bool").sum()
        was_included = hasattr(ds_result, "bool")
        self.assertTrue(
            was_included,
            msg=f"Groupby key columns were included in main data. (they should not be by default)",
        )

        # included (by keyword)
        ds_result = ds.gb("bool", include_keys=True).sum()
        was_included = hasattr(ds_result, "bool")
        self.assertTrue(
            was_included,
            msg=f"Groupby key columns were NOT included in main data, even though the keyword was set.",
        )

        # included (by global flag)
        GroupBy.include_keys = True
        ds_result = ds.gb("bool").sum()
        was_included = hasattr(ds_result, "bool")
        self.assertTrue(
            was_included,
            msg=f"Groupby key columns were NOT included in main data, even though the global flag was set.",
        )
        GroupBy.include_keys = False

    def test_ops_str_allowed(self):
        gb = ds.gb("bool")
        result_dict = {"first": [4, 3], "last": [1, 2]}
        for name, correct in result_dict.items():
            call = getattr(gb, name)
            ds_result = call()
            # check numeric columns
            for dt in arr_types:
                column = getattr(ds_result, dt.__name__)
                is_correct = bool(np.all(column == np.array(correct, dtype=dt)))
                self.assertTrue(
                    is_correct,
                    msg=f"Incorrect result for keys {ds_result.label_get_names()} in column with datatype {dt} after {name} operation.",
                )

            # check string columns
            for dt_str in arr_types_string:
                column = getattr(ds_result, dt_str.__name__).astype(np.str_)
                is_correct = bool(np.all(column == np.array(correct, dtype=np.str_)))
                self.assertTrue(
                    is_correct,
                    msg=f"Incorrect result for keys {ds_result.label_get_names()} in column with datatype {dt_str} after {name} operation.",
                )

    def test_math_ops_gb_nums_same_return(self):
        GroupBy.include_keys = True
        test_funcs = ["sum", "first", "last", "nansum"]
        for name in test_funcs:
            # check sorted
            for dt in arr_types:
                call = getattr(ds.gb(dt.__name__), name)
                ds_result = call()
                for dt2 in arr_types:
                    column = getattr(ds_result, dt2.__name__)
                    is_correct = bool(np.all(column == np.array(num_list_sorted, dtype=dt)))
                    self.assertTrue(
                        is_correct,
                        msg=f"Incorrect result for column {dt2.__name__} after grouping by {dt.__name__} for {name} operation. (sorted)",
                    )

            # check unsorted
            for dt in arr_types:
                call = getattr(ds.gbu(dt.__name__), name)
                ds_result = call()
                for dt2 in arr_types:
                    column = getattr(ds_result, dt2.__name__)
                    is_correct = bool(np.all(column == np.array(num_list, dtype=dt)))
                    self.assertTrue(
                        is_correct,
                        msg=f"Incorrect result for column {dt2.__name__} after grouping by {dt.__name__} for {name} operation. (unsorted)",
                    )
        GroupBy.include_keys = False

    # funcs that return nan on single item
    def test_math_ops_nan_return(self):
        GroupBy.include_keys = True
        test_funcs = ["std", "var", "nanstd", "nanvar"]
        for dt in arr_types + arr_types_string:
            gb = ds.gb(dt.__name__)
            # print("***nan_return\n", ds, dt.__name__)
            for func_name in test_funcs:
                call = getattr(gb, func_name)
                ds_result = call()
                for dt2 in arr_types:
                    column = getattr(ds_result, dt2.__name__)
                    # skip over what we grouped by
                    if dt.__name__ != dt2.__name__:
                        for item in column:
                            self.assertNotEqual(
                                item,
                                item,
                                msg=f"Item was not NaN for groupby {dt} in column {dt2} for {func_name} operation. Got {item} instead.",
                            )
        GroupBy.include_keys = False

    def test_count(self):
        """
        The result of count only returns a single column, which displays a count of the number of keys in that grouping.
        """
        result_ds = ds.gb("bool").count()
        self.assertEqual(result_ds._ncols, 2, msg=f"Groupby count did not return a single column.")

        correct_count = np.array([2, 3])
        result_arr = result_ds.Count
        is_correct = bool(np.all(correct_count == result_arr))
        self.assertTrue(
            is_correct,
            msg=f"Groupby bool did not product the correct result for count. returned {result_arr}, expected {correct_count}",
        )

    def test_nan_funcs(self):
        GroupBy.include_keys = True
        nan_dict = {
            "f32": np.arange(5, dtype=np.float32),
            "f64": np.arange(5, dtype=np.float64),
        }
        nan_dict["f32"][0] = np.nan
        nan_dict["f64"][0] = np.nan
        nan_ds = Dataset(nan_dict)
        # TODO: add when more nan functions have been implemented
        correct = np.array([1.0, 2.0, 3.0, 4.0, 0.0])
        nan_funcs = ["nansum"]
        col_names = ["f32", "f64"]

        for c_name in col_names:
            gb = nan_ds.gb(c_name)
            for f_name in nan_funcs:
                call = getattr(gb, f_name)
                result_ds = call()
                for result_name in col_names:
                    column = getattr(result_ds, result_name)
                    # skip groupby column
                    if c_name != result_name:
                        is_correct = bool(np.all(column == correct))
                        self.assertTrue(
                            is_correct,
                            msg=f"Incorrect result for groupby {c_name} in column {result_name} for {f_name} operation.",
                        )
        GroupBy.include_keys = False

    def test_not_supported(self):
        """
        Remove these from the following list when they are implemented.
        """
        not_implemented = [
            "sem",
            "ohlc",
            "describe",
            "expanding",
            "ngroup",
            "rank",
            "head",
            "tail",
            "stack",
            "unstack",
        ]
        # 8/28/2018 - SJK removed nth
        not_implemented_one_arg = ["resample"]
        not_implemented_two_arg = ["transform"]
        gb = ds.gb("bool")

        for func in not_implemented:
            call = getattr(gb, func)
            with self.assertRaises(
                NotImplementedError,
                msg=f"Failed to raise an error for {func}. If this HAS been implemented, remove test from list in test_not_supported in test_sfw_groupby.py",
            ):
                result = call()

        for func in not_implemented_one_arg:
            call = getattr(gb, func)
            with self.assertRaises(
                NotImplementedError,
                msg=f"Failed to raise an error for {func}. If this HAS been implemented, remove test from list in test_not_supported in test_sfw_groupby.py",
            ):
                result = call(1)

        for func in not_implemented_two_arg:
            with self.assertRaises(
                NotImplementedError,
                msg=f"Failed to raise an error for {func}. If this HAS been implemented, remove test from list in test_not_supported in test_sfw_groupby.py",
            ):
                result = call(1, 2)

    def array_equal(self, arr1, arr2):
        subr = arr1 - arr2
        sumr = sum(subr == 0)
        result = sumr == len(arr1)
        if not result:
            print("array comparison failed", arr1, arr2)
        return result

    def assertArrayEqual(self, ar1, ar2):
        assert (ar1 == ar2).all(axis=None), f"array equality failed\nar1 {ar1}\nar2 {ar2}"

    def assertArrayAlmostEqual(self, ar1, ar2, places=5):
        for v1, v2 in zip(ar1, ar2):
            self.assertAlmostEqual(v1, v2, places)

    def test_gb_categoricals(self):
        codes = [1, 44, 44, 133, 75, 75, 75, 1]
        stringlist = ["a", "b", "c", "d", "e", "e", "f", "g"]
        c1 = Categorical(codes, LikertDecision, sort_gb=True)
        c2 = Categorical(stringlist)
        d = {"nums": np.arange(8)}

        # from enum only
        d_enum = d.copy()
        d_enum["cat_from_enum"] = c1
        ds_enum = Dataset(d_enum)
        enum_result = ds_enum.gb("cat_from_enum").sum()
        correct = FastArray([3, 15, 3, 7], dtype=np.int64)
        self.assertTrue(
            self.array_equal(correct, enum_result.nums),
            msg=f"Incorrect sum when grouping by enum categorical.\nExpected {correct}\nActual {enum_result.nums}",
        )

        # from list only
        d_list = d.copy()
        d_list["cat_from_list"] = c2
        ds_list = Dataset(d_list)
        list_result = ds_list.gb("cat_from_list").sum()
        correct = FastArray([0, 1, 2, 3, 9, 6, 7], dtype=np.int64)
        self.assertTrue(
            self.array_equal(correct, list_result.nums),
            msg=f"Incorrect sum when grouping by list categorical.",
        )

        d_both = d_enum.copy()
        d_both["cat_from_list"] = c2
        ds_both = Dataset(d_both)

        # by enum, list
        result = ds_both.gb(["cat_from_enum", "cat_from_list"]).sum()
        num_result = result.nums
        correct = FastArray([0, 7, 1, 2, 9, 6, 3], dtype=np.int64)
        self.assertTrue(
            self.array_equal(correct, num_result),
            msg=f"Incorrect sum when grouping by enum, list categoricals.",
        )

        # by list, enum
        result = ds_both.gb(["cat_from_list", "cat_from_enum"]).sum()
        num_result = result.nums
        correct = FastArray([0, 1, 2, 3, 9, 6, 7], dtype=np.int64)
        self.assertTrue(
            self.array_equal(correct, num_result),
            msg=f"Incorrect sum when grouping by list, enum categoricals.",
        )

    def test_ops(self):
        ds = Dataset(
            {
                "test": arange(300000) % 3,
                "test2": arange(300000.0),
                "test2i": arange(300000),
                "test3": arange(300000) % 3,
            }
        )
        gb = ds.groupby("test")
        result = gb.mean()
        self.assertTrue(result.test2[0] == result.test2i[0])
        self.assertTrue(result.test2[1] == result.test2i[1])
        self.assertTrue(result.test3[1] == 1.0)
        result = gb.median()
        result = gb.trimbr()
        result = gb.nanmedian()
        result = gb.nanquantile(q=0.1)
        result = gb.rolling_quantile(q=0.71)
        result = gb.nanquantile(q=[0.15, 0.45, 0.97])
        result = gb.percentile(q=[0.0, 50.0, 0.6, 100.0])

    def test_projections(self):
        num_rows_trade = 1_000_000
        num_symbols = 450
        Trade_Dates = ["20180602", "20180603", "20180604", "20180605", "20180606"]
        Exchanges = np.array(["EXCH1", "EXCH2", "EXCH3"])
        np.random.seed(1234)
        ds = Dataset(
            {
                "SymbolID": np.random.randint(0, num_symbols, size=num_rows_trade),
                "Exchange": Exchanges[np.random.randint(0, Exchanges.shape[0], size=num_rows_trade)],
                "Trade_Date": [Trade_Dates[int(i * len(Trade_Dates) / num_rows_trade)] for i in range(num_rows_trade)],
                "Time": [int(i % (num_rows_trade / len(Trade_Dates))) for i in range(num_rows_trade)],
                "Price": 100 * (1.0 + 0.0005 * np.random.randn(num_rows_trade)),
                "Size": 10 * np.array(1 + 30 * np.random.rand(num_rows_trade), dtype=np.int64),
            }
        )
        num_rows_quote = 1_000_000
        ds2 = Dataset(
            {
                "SymbolID": np.random.randint(0, num_symbols, size=num_rows_quote),
                "Exchange": Exchanges[np.random.randint(0, Exchanges.shape[0], size=num_rows_quote)],
                "Trade_Date": [Trade_Dates[int(i * len(Trade_Dates) / num_rows_quote)] for i in range(num_rows_quote)],
                "Time": [int(i % (num_rows_quote / len(Trade_Dates))) for i in range(num_rows_quote)],
                "Bid": 100 * (1.0 - 0.001 + 0.0005 * np.random.randn(num_rows_quote)),
                "Ask": 100 * (1.0 + 0.001 + 0.0005 * np.random.randn(num_rows_quote)),
            }
        )
        threshold = Dataset({"Is_Below_Thresdhold": np.random.rand(num_rows_quote) < 0.75})
        trade_time = Dataset({"time_2500": (ds.Time / 2500).astype(int)})
        trades = Dataset({}).concat_columns([ds, trade_time], do_copy=False)

        # Create GroupBy and corresponding Categorical
        trade_gb = trades.groupby(["SymbolID", "Exchange", "Trade_Date", "time_2500"])
        trade_cat = Categorical([ds.SymbolID, ds.Exchange, ds.Trade_Date, trade_time.time_2500])

        # Call sum() and count()
        self.assertEqual(trade_gb.sum().shape, (455654, 7))
        self.assertEqual(trade_cat.sum(ds).shape, (455654, 7))
        self.assertEqual(trade_gb.count().shape, (455654, 5))
        # 8/24/2018 SJK - multikey categorical groupby now returns multiple columns for groupby keys
        self.assertEqual(trade_cat.count().shape, (455654, 5))
        b1 = trade_gb.count().Count.mean()
        b1c = trade_cat.count().Count.mean()
        b2 = trade_gb.count().shape[0]
        self.assertAlmostEqual(ds.shape[0], b1 * b2, places=5)
        self.assertAlmostEqual(ds.shape[0], b1c * b2, places=5)

        # Create ds augmented with filtered ID
        trade_ds = Dataset({"ID": trade_gb.grouping.ikey})
        trade_ds_below_threshold = ds * threshold.Is_Below_Thresdhold
        trade_ds_below_thresholdb = Dataset.concat_columns([trade_ds_below_threshold, trade_ds], do_copy=False)

        # Create trade_ds size projection using GroupBy
        trade_gb_id = trade_ds_below_thresholdb.groupby("ID")
        trade_sizes_ds = trade_gb_id["Size"].sum()
        trade_size_ds = trade_sizes_ds.Size[trade_ds_below_thresholdb.ID - 1]
        self.assertEqual(trade_size_ds.shape[0], ds.shape[0])

        # Create trade_ds size projection using Categorical
        trade_sizes_cat_ds = trade_cat.sum(trade_ds_below_thresholdb.Size)
        trade_size_cat_ds = trade_sizes_cat_ds.Size[trade_cat - 1]
        self.assertArrayAlmostEqual(trade_size_ds, trade_size_cat_ds, places=6)

        # Create trade_ds size projection using Pandas groupby
        ptrade_ds_below_thresholdb = dataset_as_pandas_df(trade_ds_below_thresholdb)
        ptrade_gb_id = ptrade_ds_below_thresholdb.groupby("ID")
        trade_sizes_pd_ds = ptrade_gb_id.sum()
        trade_size_pd_ds = trade_sizes_pd_ds.Size.values[ptrade_gb_id.ngroup()]
        self.assertArrayAlmostEqual(trade_size_ds, trade_size_pd_ds, places=6)

    def test_reductions(self):
        message_types = [
            "CREATE",
            "RUN",
            "CREATE",
            "RUN",
            "RUN",
            "RUN",
            "RUN",
            "CANCEL",
            "RUN",
            "RUN",
            "RUN",
            "CANCEL",
        ]
        order_ids = [1, 1, 2, 1, 2, 2, 1, 2, 2, 2, 2, 1]
        seconds = [50, 70, 72, 75, 90, 88, 95, 97, 98, 115, 116, 120]
        shares = [0, 200, 0, 500, 100, 400, 100, 0, 300, 150, 150, 0]
        d2 = dict(
            message_type=message_types,
            order_id=order_ids,
            second=seconds,
            shares=shares,
        )
        dat = Dataset(d2)
        dat = dat[["order_id", "message_type", "second", "shares"]]

        # Numeric reduction
        dsr = dat.groupby("order_id").sum()
        self.assertEqual(dsr.shape, (2, 3))
        self.assertArrayEqual(dsr.order_id, [1, 2])
        self.assertArrayEqual(dsr.second, [410, 676])
        self.assertArrayEqual(dsr.shares, [800, 1100])

        # Numeric reduction with all columns returned
        dsr = dat.groupby("order_id", return_all=True).sum()
        self.assertEqual(dsr.shape, (2, 4))
        self.assertEqual(dsr.keys()[1], "message_type")

        # Order-based reduction
        dsr = dat.groupby("order_id").first()
        self.assertEqual(dsr.shape, (2, 4))
        self.assertArrayEqual(dsr.order_id, [1, 2])
        self.assertArrayEqual(dsr.message_type, ["CREATE", "CREATE"])
        self.assertArrayEqual(dsr.second, [50, 72])
        self.assertArrayEqual(dsr.shares, [0, 0])

        # Order-based reduction, which returns all columns regardless
        dsr = dat.groupby("order_id", return_all=True).first()
        self.assertEqual(dsr.shape, (2, 4))

        # Order-based reduction with multiple keys
        dsr = dat.groupby(["order_id", "message_type"]).first()
        self.assertEqual(dsr.shape, (6, 4))
        self.assertArrayEqual(dsr.order_id, [1, 1, 1, 2, 2, 2])
        self.assertArrayEqual(dsr.message_type, ["CANCEL", "CREATE", "RUN", "CANCEL", "CREATE", "RUN"])
        self.assertArrayEqual(dsr.second, [120, 50, 70, 97, 72, 90])
        self.assertArrayEqual(dsr.shares, [0, 0, 200, 0, 0, 100])

        # On a subset of columns
        gb = dat.groupby("order_id")
        dsr = gb["shares"].sum()
        self.assertEqual(dsr.shape, (2, 2))
        self.assertArrayEqual(dsr.shares, [800, 1100])

        # Accumulating function
        dsr = dat.groupby("order_id").cumsum()
        self.assertEqual(dsr.shape, (12, 2))
        self.assertArrayEqual(dsr.shares, [0, 200, 0, 700, 100, 500, 800, 500, 800, 950, 1100, 800])

        # return_all has no effect with accumulating functions
        # 8/23/2018 SJK - changed behavior so return all shows the keys
        dsr = dat.groupby("order_id", return_all=True).cumsum()
        self.assertEqual(dsr.shape, (12, 3))

        # Add cum_shares back to a dataset
        dat["cum_shares"] = dat.groupby("order_id").shares.cumsum().shares
        self.assertEqual(dat.shape, (12, 5))
        self.assertArrayEqual(dat.cum_shares, gb.shares.cumsum().shares)

        # On a subset of columns
        dsr = dat.groupby("order_id")[["shares", "second"]].cumsum()
        self.assertEqual(dsr.shape, (12, 2))
        self.assertArrayEqual(dsr.shares, [0, 200, 0, 700, 100, 500, 800, 500, 800, 950, 1100, 800])
        self.assertArrayEqual(dsr.second, [50, 120, 72, 195, 162, 250, 290, 347, 445, 560, 676, 410])

        # On a subset of columns with a filter
        f = FastArray(
            [
                True,
                False,
                True,
                False,
                True,
                False,
                True,
                False,
                True,
                False,
                True,
                False,
            ]
        )
        dsr = dat.groupby("order_id")[["shares", "second"]].cumsum(filter=f)
        self.assertEqual(dsr.shape, (12, 2))
        self.assertArrayEqual(dsr.shares, [0, 0, 0, 0, 100, 100, 100, 100, 400, 400, 550, 100])
        self.assertArrayEqual(dsr.second, [50, 50, 72, 50, 162, 162, 145, 162, 260, 260, 376, 145])

        # On shares and second with filter at groupby construction
        dsr = dat.groupby("order_id", filter=f)[["shares", "second"]].cumsum()
        inv = INVALID_DICT[dsr.shares[0].dtype.num]
        self.assertEqual(dsr.shape, (12, 2))
        self.assertArrayEqual(dsr.shares, [0, inv, 0, inv, 100, inv, 100, inv, 400, inv, 550, inv])
        self.assertArrayEqual(dsr.second, [50, inv, 72, inv, 162, inv, 145, inv, 260, inv, 376, inv])

        # Using agg function
        dsr = gb[["second", "shares"]].agg(["sum", "mean"])
        self.assertEqual(dsr.shape, (2, 2))
        self.assertArrayEqual(dsr.Sum.second, [410, 676])
        self.assertArrayEqual(dsr.Sum.shares, [800, 1100])
        self.assertArrayAlmostEqual(dsr.Mean.second, [82.00, 96.57], places=2)
        self.assertArrayAlmostEqual(dsr.Mean.shares, [160.00, 157.14], places=2)

        # Check for issue when bracket indexing on groupby
        f = open(os.devnull, "w")
        print(gb, file=f)
        f.close()
        dsr = gb[["second", "shares"]].agg(["sum", "mean"])

        # Using different functions on different columns
        dsr = gb.agg({"second": "sum", "shares": ["max", "mean"]})
        self.assertEqual(dsr.shape, (2, 3))
        self.assertArrayEqual(dsr.Sum.second, [410, 676])
        self.assertArrayEqual(dsr.Max.shares, [500, 400])
        self.assertArrayAlmostEqual(dsr.Mean.shares, [160.00, 157.14], places=2)

        # Using numpy functions
        dsr = gb.agg({"second": np.sum, "shares": [np.max, np.mean]})
        self.assertEqual(dsr.shape, (2, 3))
        self.assertArrayEqual(dsr.Sum.second, [410, 676])
        self.assertArrayEqual(dsr.Max.shares, [500, 400])
        self.assertArrayAlmostEqual(dsr.Mean.shares, [160.00, 157.14], places=2)

        # Alternate way to add to multiset
        gb = dat.groupby("order_id")
        ms = gb[["shares"]].agg(["max", "mean"])
        ms.Sum = gb[["second"]].sum()
        self.assertEqual(ms.shape, (2, 3))
        self.assertArrayEqual(ms.Sum.second, [410, 676])
        self.assertArrayEqual(ms.Max.shares, [500, 400])
        self.assertArrayAlmostEqual(ms.Mean.shares, [160.00, 157.14], places=2)

    def test_iter(self):
        correct_keys = FastArray(["e", "d", "b", "c", "a"])
        correct_idx = [[0, 1, 4, 7], [2, 9], [3], [5, 6], [8]]
        str_arr = FastArray(["e", "e", "d", "b", "e", "c", "c", "e", "a", "d"])

        gb = Dataset({"keycol": str_arr, "idxcol": arange(10)})
        gb = gb.gb("keycol")
        for i, tup in enumerate(gb):
            self.assertEqual(tup[0], correct_keys[i])
            self.assertTrue(bool(np.all(tup[1].idxcol == correct_idx[i])))

    def test_shift(self):
        import pandas as pd

        ds = Dataset({"col_" + str(i): np.random.rand(30) for i in range(5)})
        ds.keycol = np.random.choice(["a", "b", "c"], 30)
        df = pd.DataFrame(ds.asdict())

        rt_result = ds.gb("keycol").shift(periods=-2).trim()
        pd_result = df.groupby("keycol").shift(periods=-2).dropna(axis="rows")

        for k, v in rt_result.items():
            self.assertTrue(bool(np.all(v == pd_result[k])))

        rt_result = ds.gb("keycol").shift(periods=3).trim()
        pd_result = df.groupby("keycol").shift(periods=3).dropna(axis="rows")

        for k, v in rt_result.items():
            self.assertTrue(bool(np.all(v == pd_result[k])))

    def test_diff(self):
        import pandas as pd

        ds = Dataset({"col_" + str(i): np.random.rand(10) for i in range(5)})
        ds.keycol = np.random.choice(["a", "b", "c"], 10)
        df = pd.DataFrame(ds.asdict())

        rt_result = ds.gb("keycol").rolling_diff()
        pd_result = df.groupby("keycol").diff()

        for k, v in rt_result.items():
            pdc = pd_result[k]
            pdcnan = isnan(pdc)
            self.assertTrue(bool(np.all(isnan(v) == pdcnan)), msg=f"{v} {pdc}")

            masked_valid_pd = isnotnan(pdc)
            masked_valid_rt = isnotnan(v)

            self.assertTrue(bool(np.all(masked_valid_pd == masked_valid_rt)))

    def test_pad(self):
        arrsize = 100
        numrows = 20
        ds = Dataset({"time": arange(arrsize * 1.0)})
        ds.data = np.random.randint(numrows, size=arrsize)
        ds.data2 = np.random.randint(numrows, size=arrsize)
        symbols = ["ZYGO", "YHOO", "FB", "GOOG", "IBM"]
        ds.symbol2 = Cat(1 + ds.data, list("ABCDEFGHIJKLMNOPQRST"))
        ds.symbol = Cat(1 + arange(arrsize) % len(symbols), symbols)
        ds.time[[3, 4, 7]] = nan
        newds = ds.gb("symbol").pad()
        self.assertTrue(newds.time[7] == 2.00)
        self.assertTrue(newds.time[3] != newds.time[3])
        newds = ds.gb("symbol").backfill()
        self.assertTrue(newds.time[7] == 12.00)

        # see if we can pull a group
        newds = ds.gb("symbol").get_group("YHOO")
        self.assertTrue(np.all(newds.symbol == "YHOO"))

    def test_transform(self):

        arrsize = 200
        numrows = 7

        ds = Dataset({"time": arange(arrsize * 1.0)})
        ds.data = np.random.randint(numrows, size=arrsize)
        ds.data2 = np.random.randint(numrows, size=arrsize)
        symbols = ["AAPL", "AMZN", "FB", "GOOG", "IBM"]
        ds.symbol = Cat(1 + arange(arrsize) % len(symbols), symbols)
        newds = ds.gb("symbol")["data"].sum(transform=True)

        # removed from test since gbkeys not returned in transform
        # self.assertTrue(np.all(newds.symbol == ds.symbol))
        catds = ds.symbol.sum(ds.data, transform=True)
        self.assertTrue(np.all(newds[0] == catds[0]))
        # test showfilter
        catds = ds.symbol.sum(ds.data, showfilter=True, transform=True)
        self.assertTrue(np.all(newds[0] == catds[0]))

        # test diff
        result1 = ds.gb("symbol").apply_nonreduce(TypeRegister.FastArray.diff)
        result2 = ds.gb("symbol").diff()
        self.assertTrue(result1.equals(result2))


if __name__ == "__main__":
    tester = unittest.main()



from .rt_groupby import GroupBy
from .rt_groupbyops import GroupByOps


# =====================================================================================================
# =====================================================================================================
class PGroupBy(GroupBy):
    """
    Parameters
    ----------
    dataset: Dataset
        The dataset object

    keys: list
        List of column names to groupby

    filter: array of bools
        Boolean mask array applied as filter before grouping

    return_all: bool
        Default to False. When set to True will return all the dataset columns for every operation.

    hintSize: int
        Hint size for the hash

    sort: bool
        Default to True.  Indicates


    Attributes
    ----------
    gbkeys:  dictionary of numpy arrays binned from
    isortrows: sorted index or None

    """

    DebugMode = False
    ShowEmpty = True

    TestCatGb = True

    def __init__(self, *args, **kwargs):
        super()._init(self, *args, **kwargs)

    # ---------------------------------------------------------------
    def copy(self, deep=True):
        pass



__all__ = ["normalize_zscore", "normalize_minmax"]

import warnings

import numpy as np

from .rt_enum import TypeRegister

# Routines to normalize data for machine learing


def normalize_zscore(arr):
    _mean = arr.nanmean()
    _std = arr.nanstd()
    return (arr - _mean) / _std


def normalize_minmax(arr):
    _min = arr.nanmin()
    _max = arr.nanmax()
    _range = _max - _min
    return (arr - _min) / _range


# TODO seaborn is no longer a dependency, nor imported so this will break when called
# def plot_features(ds, features, row_mask=None):
#     ''' uses seaborn lmplot '''
#     import pandas as pd
#     if row_mask is None:
#         g=ds[features]
#     else:
#         g=ds[features][row_mask,:]
#
#     g['opt']=y_train
#     g['symbol']=foo.pSym[goodsymbol_mask]
#
#     pd_data= pd.DataFrame(g.asdict())
#
#     #sns.relplot(x="date", y="opt",  col="symbol", kind="line", data=pd.DataFrame(g.asdict()))
#     for c in features:
#         sns.lmplot(x=c, y="opt",  col="symbol",  data=pd_data)
